from dataclasses import asdict, dataclass, field, Field, MISSING, is_dataclass, fields
import copy
from typing import Any, Callable, TypeVar, Type, List
from typing_extensions import dataclass_transform
from dman.utils import sjson


BASE_KEY = "__base"
DESCRIPTOR_KEY = "__descriptor"
DESCRIPTOR = Any
CONFIGFIELD = "_configclass"

_T = TypeVar("_T")

@dataclass_transform(field_specifiers=(Field, field))
def idataclass(
    cls=None,
    *,
    init=True,
    repr=True,
    eq=True,
    order=False,
    unsafe_hash=False,
    frozen=False,
) -> Callable[[Type[_T]], Type[_T]]:
    """
    Convert a class to an iterable dataclass.
        Returns the same class as was passed in, with dunder methods added based on the fields
        defined in the class.
        The class is automatically made ``iterable`` by adding ``__iter__``

        The arguments of the ``dataclass`` decorator are provided.

    :param bool init: add an ``__init__`` method.
    :param bool repr: add a ``__repr__`` method.
    :param bool order: rich comparison dunder methods are added.
    :param bool unsafe_hash: add a ``__hash__`` method function.
    :param bool frozen: fields may not be assigned to after instance creation.
    """
    def wrap(cls):
        res = dataclass(
            cls,
            init=init,
            repr=repr,
            eq=eq,
            order=order,
            unsafe_hash=unsafe_hash,
            frozen=frozen,
        )
        setattr(
            res, "__iter__", lambda self: (getattr(self, f.name) for f in fields(self))
        )
        return res

    return wrap if cls is None else wrap(cls)


def wrapfield(
    wrap: DESCRIPTOR,
    /,
    *,
    default=MISSING,
    default_factory=MISSING,
    init=True,
    repr=False,
    hash=False,
    compare=False,
    metadata: dict = None,
) -> Field:

    _metadata = {DESCRIPTOR_KEY: wrap}
    if metadata is not None:
        _metadata[BASE_KEY] = metadata
    if not hasattr(wrap, "__set__"):
        init = False

    return field(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=_metadata,
    )


def is_wrapfield(fld: Field):
    return fld.metadata is not None and fld.metadata.get(DESCRIPTOR_KEY, None) is not None


def get_descriptor(fld: Field):
    return fld.metadata.get(DESCRIPTOR_KEY, None)


def set_descriptor(fld: Field, value):
    fld.metadata[DESCRIPTOR_KEY] = value


@dataclass_transform(field_specifiers=(wrapfield, field, Field))
def wrappedclass(
    cls=None,
    /,
    *,
    init=True,
    repr=True,
    eq=True,
    order=False,
    unsafe_hash=False,
    frozen=False,
) -> Callable[[Type[_T]], Type[_T]]:
    def wrap(cls):
        # convert to dataclass
        res = cls
        if not is_dataclass(cls):
            res = dataclass(
                cls,
                init=init,
                repr=repr,
                eq=eq,
                order=order,
                unsafe_hash=unsafe_hash,
                frozen=frozen,
            )

        # go through fields and process wrapped fields
        for f in fields(res):
            if is_wrapfield(f):
                prop = f.metadata.get(DESCRIPTOR_KEY)
                if hasattr(prop, "__set_name__"):
                    prop.__set_name__(cls, f.name)
                f.metadata = f.metadata.get(BASE_KEY, None)
                setattr(res, f.name, prop)

        return res

    return wrap if cls is None else wrap(cls)


def is_configclass(cls):
    return hasattr(cls, "_configclass")


class OptionField:
    def __init__(self, tp: Type = None, options: List = None):
        self.tp = tp
        self.options = options
        self.key = None

    def __set_name__(self, owner, name):
        self.key = f"_{name}"

    def __get__(self, obj, objtype=None):
        if not hasattr(obj, self.key):
            return self.options
        return getattr(obj, self.key)

    def __set__(self, obj, value):
        if not isinstance(value, self.tp):
            raise ValueError(
                f"Invalid type assigned to field. Expected {self.tp} got {type(value)}"
            )
        if self.options is not None and value not in self.options:
            raise ValueError(
                (
                    f"Invalid value assigned to field. "
                    f'Expected "{"|".join([str(s) for s in  self.options])}" got "{value}".'
                )
            )
        setattr(obj, self.key, value)


def optionfield(
    options: list = None,
    *,
    default=MISSING,
    default_factory=MISSING,
    repr: bool = True,
    tp: Type = None,
):
    wrap = OptionField(tp=tp, options=options)
    return wrapfield(
        wrap,
        default=default,
        default_factory=default_factory,
        repr=repr,
    )


@dataclass_transform(field_specifiers=(wrapfield, optionfield, field, OptionField, Field))
def configclass(cls=None, /, *, atomic: bool = True, force_defaults: bool = True) -> Callable[[Type[_T]], Type[_T]]:
    """Convert a class to a configclass."""

    def wrap(cls):
        if not is_dataclass(cls):
            cls = dataclass(cls)

        for f in fields(cls):
            # get the list of options (if available)
            options = cls.__dict__.get(f.name, None)
            options = None if not isinstance(options, OptionField) else options.options

            # check if the type is atomic if required
            if atomic and (
                f.type not in sjson.atomic_types and not is_configclass(f.type)
            ):
                raise ValueError(
                    f"Specified field with non-atomic type {f.type} for atomic configclass."
                )

            # update field with new type
            wrapped_field: Field = optionfield(tp=f.type, options=options)
            f.metadata = wrapped_field.metadata

            # check if all types have defaults
            if force_defaults and f.default is MISSING and f.default_factory is MISSING:
                raise ValueError(
                    (
                        f"All fields for the configclass should have defaults."
                        f" Found {f.name} without one."
                    )
                )

        cls = wrappedclass(cls)
        setattr(cls, CONFIGFIELD, True)
        return cls

    return wrap if cls is None else wrap(cls)


def override(left, right, *, inplace: bool = False, ignore: list = None):
    if not inplace:
        left = copy.deepcopy(left)
    if ignore is None:
        ignore = []

    if is_configclass(right):
        right = asdict(right)
    for f in fields(left):
        value = right.get(f.name, MISSING)
        if value is MISSING or value in ignore:
            continue
        if is_configclass(f.type):
            setattr(left, f.name, override(getattr(left, f.name), value, ignore=ignore))
        else:
            setattr(left, f.name, value)

    return left



"""
High-level saving and loading functionality for ``dman``.
"""


from contextlib import contextmanager
from dataclasses import MISSING
import os
from tempfile import TemporaryDirectory
from uuid import uuid4
from pathlib import Path
import shutil

from dman.core import log
from dman.model.record import Context, is_removable, record, remove
from dman.core.serializables import deserialize, is_serializable, serialize
from dman.core.storables import is_storable
from dman.utils import sjson
from dman.core.path import normalize_path, Target, AUTO, gitignore
from dman.core.storables import storable

import signal


class _InterruptTracker:
    def __init__(self):
        self.value = None
    
    def handler(self, signum, frame):
        self.value = (signum, frame)


@contextmanager
def uninterrupted():
    """Return context that suppresses keyboard interrupts until it exists."""
    tracker = _InterruptTracker()
    original = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, tracker.handler)
    yield
    signal.signal(signal.SIGINT, original)
    if tracker.value is not None:
        tracker.value = None
        raise KeyboardInterrupt()


@contextmanager
def context(
    key: str,
    *,
    subdir: os.PathLike = "",
    cluster: bool = True,
    verbose: int = None,
    generator: str = None,
    base: os.PathLike = None,
    gitignore: bool = True,
):
    """Get a context from a mount point.
        The path of the file is determined as described below.

            If the files are clustered then the path is ``<base>/<generator>/<subdir>/<key>/<key>.<ext>``
            If cluster is set to False then the path is ``<base>/<generator>/<subdir>/<key>.<ext>``

            When base is not provided then it is set to .dman if
            it does not exist an exception is raised.

            When generator is not provided it will automatically be set based on
            the location of the script relative to the .dman folder
            (again raising an exception if it is not found). For example
            if the script is located in ``<project-root>/examples/folder/script.py``
            and .dman is located in ``<project-root>/.dman``.
            Then generator is set to cache/examples:folder:script (i.e.
            the / is replaced by : in the output).

    Args:
        key (str):  Key for the file.
        subdir (os.PathLike, optional): Specifies optional subdirectory in generator folder. Defaults to "".
        cluster (bool, optional): A subfolder ``key`` is automatically created when set to True. Defaults to True.
        verbose (bool, optional): Level of verbosity. Defaults to False
        generator (str, optional): Specifies the generator that created the file. Defaults to script label.
        base (os.PathLike, optional): Specifies the root folder. Defaults to ".dman".
        gitignore (bool, optional): Specifies whether files added to this mount point should be ignored.
    """
    with log.logger_context(level=verbose):
        ctx = Context.mount(
            key,
            subdir=subdir,
            cluster=cluster,
            generator=generator,
            base=base,
            gitignore=gitignore,
        )
        yield ctx
        ctx.close()


def store(
    key: str,
    obj,
    *,
    subdir: os.PathLike = "",
    cluster: bool = False,
    verbose: int = None,
    gitignore: bool = True,
    generator: str = None,
    base: os.PathLike = None,
):
    """
    Save a storable object.
        The path of the file is determined as described below.

            If the files are clustered then the path is ``<base>/<generator>/<subdir>/<key>/<key>.<ext>``
            If cluster is set to False then the path is ``<base>/<generator>/<subdir>/<key>.<ext>``

            When base is not provided then it is set to .dman if
            it does not exist an exception is raised.

            When generator is not provided it will automatically be set based on
            the location of the script relative to the .dman folder
            (again raising an exception if it is not found). For example
            if the script is located in ``<project-root>/examples/folder/script.py``
            and .dman is located in ``<project-root>/.dman``.
            Then generator is set to cache/examples:folder:script (i.e.
            the / is replaced by : in the output).

    Args:
        key (str):  Key for the file.
        obj (Any): The storable to store.
        subdir (os.PathLike, optional): Specifies optional subdirectory in generator folder. Defaults to "".
        cluster (bool, optional): A subfolder ``key`` is automatically created when set to True. Defaults to True.
        verbose (bool, optional): Level of verbosity. Defaults to False
        generator (str, optional): Specifies the generator that created the file. Defaults to script label.
        base (os.PathLike, optional): Specifies the root folder. Defaults to ".dman".
        gitignore (bool, optional): Specifies whether files added to this mount point should be ignored.
    """
    if not is_storable(obj):
        raise ValueError("Can only store storable objects.")

    with context(
        key,
        subdir=subdir,
        cluster=cluster,
        generator=generator,
        base=base,
        verbose=verbose,
        gitignore=gitignore,
    ) as ctx:
        rec = record(obj, stem=key)
        with log.layer(key, "storing", prefix="key"):
            path = os.path.join(ctx.directory, rec.target)
            log.emphasize(
                f'storing {type(obj).__name__} with key {key} to "{path}".', "store"
            )
            ser = serialize(rec, context=ctx)
            log.emphasize(
                f'finished storing {type(obj).__name__} with key {key} to "{path}".',
                "store",
            )
            return ser


def save(
    key: str,
    obj,
    *,
    subdir: os.PathLike = "",
    cluster: bool = True,
    verbose: int = None,
    validate: bool = None,
    gitignore: bool = True,
    generator: str = None,
    base: os.PathLike = None,
):
    """
    Save a serializable object to a file.
        The path of the file is determined as described below.

            If the files are clustered then the path is ``<base>/<generator>/<subdir>/<key>/<key>.<ext>``
            If cluster is set to False then the path is ``<base>/<generator>/<subdir>/<key>.<ext>``

            When base is not provided then it is set to .dman if
            it does not exist an exception is raised.

            When generator is not provided it will automatically be set based on
            the location of the script relative to the .dman folder
            (again raising an exception if it is not found). For example
            if the script is located in ``<project-root>/examples/folder/script.py``
            and .dman is located in ``<project-root>/.dman``.
            Then generator is set to cache/examples:folder:script (i.e.
            the / is replaced by : in the output).

    Args:
        key (str):  Key for the file.
        obj (Any): The serializable to save.
        subdir (os.PathLike, optional): Specifies optional subdirectory in generator folder. Defaults to "".
        cluster (bool, optional): A subfolder ``key`` is automatically created when set to True. Defaults to True.
        verbose (bool, optional): Level of verbosity. Defaults to False
        generator (str, optional): Specifies the generator that created the file. Defaults to script label.
        base (os.PathLike, optional): Specifies the root folder. Defaults to ".dman".
        gitignore (bool, optional): Specifies whether files added to this mount point should be ignored.
    """

    if not is_serializable(obj):
        if is_storable(obj):
            obj = record(obj)
        else:
            raise ValueError("Can only save serializable objects.")

    with context(
        key,
        subdir=subdir,
        cluster=cluster,
        generator=generator,
        base=base,
        verbose=verbose,
        gitignore=gitignore,
    ) as ctx:
        with log.layer(key, "saving", prefix="key"):
            _, target = ctx.prepare(Target(stem=key, suffix='.json'))
            path = os.path.join(ctx.directory, target)
            log.emphasize(
                f'saving {type(obj).__name__} with key "{key}" to "{normalize_path(path)}".',
                "save",
            )
            ser = serialize(obj, context=ctx)
            with open(path, "w") as f:
                sjson.dump(ser, f, indent=4)
            log.emphasize(
                f'finished saving {type(obj).__name__} with key "{key}" to "{normalize_path(path)}".',
                "save",
            )
            return ser


def load(
    key: str,
    *,
    default=MISSING,
    default_factory=MISSING,
    subdir: os.PathLike = "",
    cluster: bool = True,
    verbose: int = None,
    gitignore: bool = True,
    generator: str = None,
    base: os.PathLike = None,
):
    """
    Load a serializable or storable object from a file.
        A default value can be provided, which is returned when no file is found.
        Similarly default_factory is a function without arguments that
        is called instead of default. If both default and default_factory
        are specified then the value in default is returned.

        The path of the file is determined as described below.

            If the files are clustered then the path is ``<base>/<generator>/<subdir>/<key>/<key>.<ext>``
            If cluster is set to False then the path is ``<base>/<generator>/<subdir>/<key>.<ext>``

            When base is not provided then it is set to .dman if
            it does not exist an exception is raised.

            When generator is not provided it will automatically be set based on
            the location of the script relative to the .dman folder
            (again raising an exception if it is not found). For example
            if the script is located in ``<project-root>/examples/folder/script.py``
            and .dman is located in ``<project-root>/.dman``.
            Then generator is set to cache/examples:folder:script (i.e.
            the / is replaced by : in the output).

    Args:
        key (str):  Key for the file.
        default (Any, optional): The default value.
        default_factory (Callable, optional): Method with no argument that produces the default value.
        subdir (os.PathLike, optional): Specifies optional subdirectory in generator folder. Defaults to "".
        cluster (bool, optional): A subfolder ``key`` is automatically created when set to True. Defaults to True.
        verbose (bool, optional): Level of verbosity. Defaults to False
        generator (str, optional): Specifies the generator that created the file. Defaults to script label.
        base (os.PathLike, optional): Specifies the root folder. Defaults to ".dman".
        gitignore (bool, optional): Specifies whether files added to this mount point should be ignored.
    
    Returns:
        Loaded object or default value if file does not exist.
    """

    with context(
        key,
        subdir=subdir,
        cluster=cluster,
        generator=generator,
        base=base,
        verbose=verbose,
        gitignore=gitignore,
    ) as ctx:
        path = os.path.join(ctx.directory, key + ".json")
        with log.layer(key, "loading", prefix="key"):
            if not os.path.exists(path):
                log.emphasize(
                    f'file not available at "{normalize_path(path)}", using default',
                    "load",
                )
                if default is MISSING and default_factory is MISSING:
                    raise FileNotFoundError(
                        f'could not find tracked file "{normalize_path(path)}".'
                    )
                elif default is MISSING:
                    return default_factory()
                else:
                    return default

            log.emphasize(
                f'loading with key "{key}" from "{normalize_path(path)}".', "load"
            )
            with open(path, "r") as f:
                ser = sjson.load(f)
            res = deserialize(ser, context=ctx)
            log.emphasize(
                f'finished loading with key "{key}" from "{normalize_path(path)}".',
                "load",
            )
            return res


def clean(
    key: str,
    *,
    subdir: os.PathLike = "",
    cluster: bool = True,
    verbose: int = None,
    generator: str = None,
    base: os.PathLike = None,
):
    """
    Remove a serializable object.
        The path of the file is determined as described below.

            If the files are clustered then the path is ``<base>/<generator>/<subdir>/<key>/<key>.<ext>``
            If cluster is set to False then the path is ``<base>/<generator>/<subdir>/<key>.<ext>``

            When base is not provided then it is set to .dman if
            it does not exist an exception is raised.

            When generator is not provided it will automatically be set based on
            the location of the script relative to the .dman folder
            (again raising an exception if it is not found). For example
            if the script is located in ``<project-root>/examples/folder/script.py``
            and .dman is located in ``<project-root>/.dman``.
            Then generator is set to cache/examples:folder:script (i.e.
            the / is replaced by : in the output).

    Args:
        key (str):  Key for the file.
        subdir (os.PathLike, optional): Specifies optional subdirectory in generator folder. Defaults to "".
        cluster (bool, optional): A subfolder ``key`` is automatically created when set to True. Defaults to True.
        verbose (bool, optional): Level of verbosity. Defaults to False
        generator (str, optional): Specifies the generator that created the file. Defaults to script label.
        base (os.PathLike, optional): Specifies the root folder. Defaults to ".dman".
    """
    obj = load(key, subdir=subdir, cluster=cluster, generator=generator, base=base, verbose=verbose, default=None)
    if obj is None:
        return
    if not is_removable(obj):
        return

    with context(
        key,
        subdir=subdir,
        cluster=cluster,
        generator=generator,
        base=base,
        verbose=verbose
    ) as ctx:
        with log.layer(key, "removing", prefix="key"):
            remove(obj, context=ctx)
            path = os.path.join(ctx.directory, key + ".json")
            if os.path.exists(path) and not os.path.isdir(path):
                os.remove(path)

    if cluster:
        if len(os.listdir(ctx.directory)) == 0:
            os.rmdir(ctx.directory)
            gitignore(Path(ctx.directory).parent, [], check=[key])
    else:
        gitignore(ctx.directory, [], check=[key+'.json'])


class Track:
    def __init__(
        self,
        key: str,
        default,
        default_factory,
        subdir: os.PathLike,
        cluster: bool,
        verbose: int,
        gitignore: bool,
        generator: str,
        base: os.PathLike,
    ) -> None:
        self.key = key
        self._content = None
        self.default = default
        self.default_factory = default_factory
        self.subdir = subdir
        self.cluster = cluster
        self.gitignore = gitignore
        self.generator = generator
        self.base = base
        self.verbose = verbose

    @property
    def content(self):
        if self._content is None:
            self.load()
        return self._content

    def save(self, unload: bool = False):
        save(
            self.key,
            self._content,
            subdir=self.subdir,
            gitignore=self.gitignore,
            generator=self.generator,
            base=self.base,
            cluster=self.cluster,
            verbose=self.verbose,
        )
        if unload:
            return self.load()

    def load(self):
        self._content = load(
            self.key,
            default=self.default,
            default_factory=self.default_factory,
            subdir=self.subdir,
            generator=self.generator,
            base=self.base,
            cluster=self.cluster,
            gitignore=self.gitignore,
            verbose=self.verbose,
        )
        return self._content

    def __enter__(self):
        return self.load()

    def __exit__(self, *_):
        return self.save()


def track(
    key: str,
    *,
    default=MISSING,
    default_factory=MISSING,
    subdir: os.PathLike = "",
    cluster: bool = True,
    verbose: int = None,
    gitignore: bool = True,
    generator: str = None,
    base: os.PathLike = None,
):
    """
    Create track a serializable or storable object with a file.
        Ideally track is used as a context: i.e. ``with track(...) as obj: ...``.
        When the context is entered the object is loaded from a file
        (or default value is returned as described below). When the context
        exits, then the file is saved.

        If the object is storable, it will automatically be wrapped in a
        record before serialization.

        The path of the file is determined as described below.

            If the files are clustered then the path is ``<base>/<generator>/<subdir>/<key>/<key>.<ext>``
            If cluster is set to False then the path is ``<base>/<generator>/<subdir>/<key>.<ext>``

            When base is not provided then it is set to .dman if
            it does not exist an exception is raised.

            When generator is not provided it will automatically be set based on
            the location of the script relative to the .dman folder
            (again raising an exception if it is not found). For example
            if the script is located in ``<project-root>/examples/folder/script.py``
            and .dman is located in ``<project-root>/.dman``.
            Then generator is set to cache/examples:folder:script (i.e.
            the / is replaced by : in the output).

    Args:
        key (str):  Key for the file.
        default (Any, optional): The default value.
        default_factory (Callable, optional): Method with no argument that produces the default value.
        subdir (os.PathLike, optional): Specifies optional subdirectory in generator folder. Defaults to "".
        cluster (bool, optional): A subfolder ``key`` is automatically created when set to True. Defaults to True.
        verbose (bool, optional): Level of verbosity. Defaults to False
        generator (str, optional): Specifies the generator that created the file. Defaults to script label.
        base (os.PathLike, optional): Specifies the root folder. Defaults to ".dman".
        gitignore (bool, optional): Specifies whether files added to this mount point should be ignored.
    """
    return Track(
        key,
        default,
        default_factory,
        subdir,
        cluster,
        verbose,
        gitignore,
        generator,
        base,
    )


from dataclasses import MISSING, asdict
import logging as backend
from contextlib import contextmanager
import os
import io as _io
import shutil
from tempfile import TemporaryDirectory
import textwrap
from tkinter import NONE
from types import ModuleType
from typing import List, Type
import re
from uuid import uuid4
from dman.core.errors import Trace


LOGGER_NAME = "dman"
DEFAULT_FORMAT = "%(indent)s%(label)s%(message)s"
DEFAULT_FORMAT_LEVEL = "%(asctime)s %(levelname)s: %(indent)s%(label)s%(message)s"
BASE_INDENT = "  "
CAPITALIZE_LEVELNAME = False

from logging import CRITICAL, FATAL, ERROR, WARN, WARNING, INFO, DEBUG, NOTSET


class IndentedFormatter(backend.Formatter):
    """Formatter that supports indentation and label specification.

    New formatting options: `indent` and `label`
    
    Example:

    >>> fmt = IndentedFormatter(fmt="%(levelname)s: %(indent)s%(label)s%(message)s")
    >>> logger = Logger('example', level=INFO)
    >>> h = backend.StreamHandler()
    >>> h.setFormatter(formatter)
    >>> log.addHandler(hdlr)
    >>> with log.layer('value', 'label'):
    ...     log.info('test')
    ...     log.info('test', 'label')
    <label type=value>
    test
    [label]: test
    <end label type=value>
    """

    def __init__(self, fmt=DEFAULT_FORMAT, datefmt=None, style="%", validate=True, capitalize_levelname: bool = False):        
        super().__init__(fmt, datefmt, style, validate)
        self.capitalize_levelname = capitalize_levelname
    
    def _format_inner(self, record: backend.LogRecord, fmt: str, include_stack: bool = False):
        if len(fmt or '') == 0: 
            return fmt
        if not include_stack:
            formatter = backend.Formatter(fmt)
            record.message = record.getMessage()
            if formatter.usesTime():
                record.asctime = self.formatTime(record, self.datefmt)
            return formatter.formatMessage(record)
        try:
            return backend.Formatter(fmt).format(record)
        except ValueError:
            return fmt
    
    def format(self, record):
        if not hasattr(record, 'label'):
            setattr(record, 'label', '')
        if not hasattr(record, 'indent'):
            setattr(record, 'indent', '')
        if self.capitalize_levelname:
            record.levelname = record.levelname.upper()
        else:
            record.levelname = record.levelname.lower()
        splits = re.split('\%\(indent\)[-0-9]*s', self._fmt)
        if len(splits) != 2 or len(record.indent) == 0:
            return super().format(record)
        prefix, indented = [self._format_inner(record, s, '%(message)' in s) for s in splits]
        lines = indented.split('\n')
        s = prefix + textwrap.indent(lines.pop(0), record.indent)
        if len(lines) > 0:
            s += '\n' + textwrap.indent('\n'.join(lines), ' '*len(prefix) + record.indent)
            
        # record.msg = s
        # record.indent = ''
        # record.label = ''
        return s

def format_type(obj):
    """Get string label for type."""
    if isinstance(obj, (ModuleType, Type)):
        return obj.__name__
    return type(obj).__name__


def default_formatter(fmt: str = DEFAULT_FORMAT, datefmt: str = None, capitalize_levelname: bool = False):
    """Return the default ``dman`` formatter.

    Args:
        fmt (str, optional): Format string.
        datefmt (str, optional): Date format specification. Defaults to None.
        capitalize_levelname (bool, optional): Capitalize the log-level name. Defaults to False.
    """
    return IndentedFormatter(fmt, datefmt, capitalize_levelname=capitalize_levelname)


try:
    from rich.logging import RichHandler
    from rich.theme import Theme
    from rich.console import Console
    from rich.text import Text
    from rich.highlighter import RegexHighlighter, Highlighter
    import rich.traceback as _rich_tb
    from rich.table import Table

    CAPITALIZE_LEVELNAME = True

    log_theme = Theme(
        {
            "backend.label": "bright_green",
            "backend.tag": "purple",
            "backend.str": "green",
            "backend.path": "green",
            "backend.filename": "green",
            "backend.error": "red",
            "backend.fail": "red",
            "backend.warning": "yellow",
            "backend.warn": "yellow",
            "backend.debug": "bright_black",
            "backend.emphasis": "blue",
            "backend.io": "bright_cyan",
        }
    )

    class LoggingHighlighter(RegexHighlighter):
        """Apply coloring to ``dman`` log messages. Highlights label, headers, paths and strings."""

        base_style = "backend."
        highlights = [
            r"(?P<label>\[(.*?)\]:)",
            r"(?<![\\\w])(?P<str>b?'''.*?(?<!\\)'''|b?'.*?(?<!\\)'|b?\"\"\".*?(?<!\\)\"\"\"|b?\".*?(?<!\\)\")",
            r"(?P<path>\B(/[-\w._:+]+)*\/)(?P<filename>[-\w._+]*)?",
            r"(?P<tag><(.*?)>)",
        ]

    class MinimalHighlighter(RegexHighlighter):
        """Only highlight label."""

        base_style = "backend."
        highlights = [
            r"(?P<label>\[(.*?)\]:)",
        ]

    class ColorHighlighter(Highlighter):
        """Highlights all basic text with specified color, still applying base to get specific colors
        for other pieces of text."""
        def __init__(self, base_color=None, base=LoggingHighlighter) -> None:
            self.base_color = base_color
            self.base = base
            super().__init__()

        def highlight(self, text):
            if self.base_color is not None:
                text.stylize(self.base_color)
            highlighter = self.base()
            highlighter.highlight(text)
            return text

    def get_highlighter(color: str, minimal: bool):
        """Get a highlighter.

        Args:
            color (str): The color of the base text.
            minimal (bool): When set to False, :class:`LoggingHighlighter` is 
                applied to the string after setting the base color.
        """
        return (LoggingHighlighter()
            if color is None
            else ColorHighlighter(
                color, base=MinimalHighlighter if minimal else LoggingHighlighter
            )
        )

    def trace2rich(trace: Trace):
        """Convert traceback to internal representation used by rich."""
        stacks = []
        for stack in trace.stacks:
            frames = []
            for frame in stack.frames:
                frames.append(_rich_tb.Frame(**asdict(frame)))
            kwargs = asdict(stack)
            kwargs.update({"frames": frames})
            stacks.append(_rich_tb.Stack(**kwargs))
        return _rich_tb.Trace(stacks) 
    
    class DManHandler(RichHandler):
        """The default logging handler used by ``dman``."""
        def _traceback_kwargs(self):
            return dict(
                width=self.tracebacks_width,
                extra_lines=self.tracebacks_extra_lines,
                theme=self.tracebacks_theme,
                word_wrap=self.tracebacks_word_wrap,
                show_locals=self.tracebacks_show_locals,
                locals_max_length=self.locals_max_length,
                locals_max_string=self.locals_max_string,
                suppress=self.tracebacks_suppress,
            )

        def emit(self, record: backend.LogRecord) -> None:
            if hasattr(record, 'trace'):
                if self.rich_tracebacks:
                    record.exc_text = ''
                else:
                    delattr(record, 'trace')
            return super().emit(record)
            
        def render(self, *, record: backend.LogRecord, traceback, message_renderable):
            if hasattr(record, 'trace'):
                traceback = _rich_tb.Traceback(
                    trace2rich(record.trace),
                    **self._traceback_kwargs()
                )

            # highlight traceback
            if isinstance(traceback, Text):
                if record.levelno == backend.WARNING:
                    traceback = ColorHighlighter('backend.warning', base=MinimalHighlighter)(traceback)
                elif record.levelno in [backend.ERROR, backend.CRITICAL]:
                    traceback = ColorHighlighter('backend.error', base=MinimalHighlighter)(traceback)
                else:
                    traceback = self.highlighter(traceback)
            
            # apply indent if necessary
            fmt = self.formatter if self.formatter else backend.Formatter()
            if len(record.indent) > 0 and traceback:
                splits = re.split('\%\(indent\)[-0-9]*s', fmt._fmt)
                if len(splits) == 2 and '%(message)' in splits[1]:
                    output = Table.grid(padding=(0, 0))
                    output.add_column()
                    output.add_column()
                    output.add_row(
                        Text(record.indent), traceback
                    )
                    traceback = output
                    
            return super().render(record=record, traceback=traceback, message_renderable=message_renderable)
    
    def default_handler(stream=None, use_rich: bool = True, **kwargs):
        """Get the default handler used by ``dman``.

        The stream is passed to the handler, which is either a standard :class:`backend.StreamHandler`
        or a :class:`DManHandler` instance. Any other keyword arguments are passed to the 
        ``__init__`` method of the handler.

        Additionally the option ``console_style`` can be set to a dictionary. 
        The keyword arguments contained within are passed to the :class:`rich.console.Console`
        initializer and the output console is passed to the `DManHandler`. 
        """
        if not use_rich:
            return backend.StreamHandler(stream)
        console_style = {'theme': log_theme, 'file': stream}
        console_style.update(kwargs.pop('console_style', {}))
        default = dict(
            rich_tracebacks=True,
            tracebacks_show_locals=False,
            console=Console(**console_style),
            enable_link_path=False
        )
        default.update(kwargs)
        return DManHandler(**default)


except ImportError as e:
    def get_highlighter(color: str, minimal: bool):
        """Get a highlighter when ``rich`` is not available. 
        In this case the options have no effect.
        """
        return None

    def default_handler(stream=NONE, **kwargs):
        """Get the default handler when ``rich`` is not available. 
        In this case the returned handler is a standard stream handler."""
        return backend.StreamHandler(stream)


def default_config(level: int = None):
    """Load the default logger configuration."""
    config(level=level, force=True)


def config(*, level=None, filename: str=None, filemode: str = 'a', stream=None, format: str = None, datefmt: str = None, handlers: List[backend.Handler] = None, force: bool = False, **kwargs):
    """Configure the ``dman`` logger.
    
    This function does nothing if the ``dman`` logger already has handlers configured,
    unless if the ``force`` keyword argument is set to true.

    The default behavior is to create a :class:`DmanHandler` using 
    the :func:`default_handler` method. Any keyword arguments not specified below are 
    passed to that method. 

    Args:
        level (_type_, optional): Set the logger level to the specified level.
        filename (str, optional): Specifies that a FileHandler be created, using the specified
            filename, rather than a StreamHandler.
        filemode (str, optional): Specifies the mode to open the file, if filename is specified
            (if filemode is unspecified, it defaults to 'a').
        stream (optional): Use the specified stream to initialize the StreamHandler. Note
            that this argument is incompatible with 'filename' - if both
            are present, 'stream' is ignored.
        format (str, optional): Use the specified format string for the handler.
        datefmt (str, optional): Use the specified date/time format.
        handlers (List[backend.Handler], optional): If specified, this should be 
            an iterable of already created handlers, which will be added to the 
            root handler. Any handler in the list which does not have a formatter 
            assigned will be assigned the formatter created in this function.
        force (bool, optional): If this keyword is specified as true, any 
            existing handlers attached to the root logger are removed and closed, 
            before carrying out the configuration as specified by the other
            arguments.. Defaults to False.

    Note that you could specify a stream created using open(filename, mode)
    rather than passing the filename and mode in. However, it should be
    remembered that StreamHandler does not close its stream (since it may be
    using sys.stdout or sys.stderr), whereas FileHandler closes its stream
    when the handler is closed.
    """
    backend.basicConfig
    if force:
        for h in logger.handlers[:]:
            logger.removeHandler(h)
            h.close()

    if len(logger.handlers) == 0:
        if handlers is None:
            if stream is not None and filename is not None:
                raise ValueError("'stream' and 'filename' should not be "
                                    "specified together")
        elif stream is not None or filename is not None:
            raise ValueError("'stream' or 'filename' should not be "
                                "specified together with 'handlers'")
        if handlers is None:
            if filename:
                h = backend.FileHandler(filename, filemode)
            else:
                h = default_handler(stream=stream, **kwargs)
            handlers = [h]

        if format is None:
            format = DEFAULT_FORMAT if kwargs.get('use_rich', True) else DEFAULT_FORMAT_LEVEL
            
        fmt = default_formatter(format, datefmt, capitalize_levelname=CAPITALIZE_LEVELNAME)
        for h in handlers:
            if h.formatter is None:
                h.setFormatter(fmt)
            logger.addHandler(h)
        
    if level is not None:
        logger.setLevel(level)

            
class Logger(backend.Logger):
    """Logger instance used by ``dman``. Provides some additional logging methods."""

    def __init__(self, name: str, level=backend.NOTSET, stack: list = None):
        super().__init__(name, level)
        self.stack = [] if stack is None else stack

    @property
    def indent(self):
        return len(self.stack)

    def format_stack(self):
        if len(self.stack) == 0:
            return ''
        return "".join([format_type(a) + "." for a in self.stack if a is not None])[:-1]

    def pack(
        self,
        msg: str,
        label: str,
        exc_info = MISSING,
        *,
        color: str = None,
        minimal: bool = False,
        use_rich_highlighter: bool = False,
    ):
        enabled = self.isEnabledFor(backend.INFO)
        stack = self.stack
        if enabled or len(stack) == 0:
            label = "" if label is None else f"[{label}]: "
        else:
            stack = self.format_stack()
            label = f"[@{stack}" + ("" if label is None else f" | {label}") + "]: "

        indent = ""
        if enabled and self.indent > 0:
            indent = BASE_INDENT * self.indent + " "

        extra = {"label": label, "indent": indent}
        if not use_rich_highlighter:
            extra["highlighter"] = get_highlighter(color, minimal)

        if exc_info is MISSING:
            return msg, extra
        if isinstance(exc_info, Trace):
            extra['trace'] = exc_info
            return msg, extra, None
        return msg, extra, exc_info

    def info(
        self,
        msg: str,
        label: str = None,
        color: str = None,
        use_rich_highlighter: bool = False,
        *args,
        **kwargs
    ):
        """Log an info message.

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
            color (str, optional): The color of the text. Defaults to None.
            use_rich_highlighter (bool, optional): Use rich highlighting. Defaults to False.
        """
        msg, extra = self.pack(msg, label, color=color, use_rich_highlighter=use_rich_highlighter)
        super().info(msg, extra=extra, *args, **kwargs)

    def debug(self, msg: str, label: str = None, *args, **kwargs):
        """Log a debug message.

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
        """
        msg, extra = self.pack(msg, label)
        super().debug(msg, extra=extra, color="backend.debug", *args, **kwargs)

    def warning(self, msg: str, label: str = None, exc_info=False, *args, **kwargs):
        """Log a warning message

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
            exc_info (bool, optional): Add exception info. Defaults to False.
        """
        msg, extra, exc_info = self.pack(msg, label, exc_info, color="backend.warning")
        super().warning(msg, extra=extra, exc_info=exc_info, *args, **kwargs)

    def error(self, msg: str, label: str = None, exc_info=False, *args, **kwargs):
        """Log an error message

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
            exc_info (bool, optional): Add exception info. Defaults to False.
        """
        msg, extra, exc_info = self.pack(msg, label, exc_info, color="backend.error")
        super().error(msg, extra=extra, exc_info=exc_info, *args, **kwargs)

    def exception(self, msg: str, label: str = None, exc_info=True, *args, **kwargs):
        """Log an exception

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
            exc_info (bool, optional): Add exception info. Defaults to True.
        """
        kwargs.update({'stacklevel': kwargs.get('stacklevel', 1)+1})
        self.error(msg, label, exc_info, *args, **kwargs)

    def emphasize(self, msg: str, label: str = None, *args, **kwargs):
        """Log an emphasized info message.

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
        """
        kwargs.update({'stacklevel': kwargs.get('stacklevel', 1)+1})
        self.info(msg, label, color="backend.emphasis", *args, **kwargs)

    def io(self, msg: str, label: str = None, *args, **kwargs):
        """Log an info message associated with io.

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
        """
        kwargs.update({'stacklevel': kwargs.get('stacklevel', 1)+1})
        self.info(msg, label, color="backend.io", *args, **kwargs)

    def header(self, msg: str, label: str, prefix: str = "type", *args, **kwargs):
        """Log an info header.

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
            prefix (str, optional): Prefix added before the message. Defaults to "type".
        """
        kwargs.update({'stacklevel': kwargs.get('stacklevel', 1)+1})
        if label is not None:
            msg = f"<{label} {prefix}={msg}>"
        self.info(msg, *args, **kwargs)

    @contextmanager
    def layer(
        self, msg: str, label: str = None, prefix: str = "type", owner: str = None, *args, **kwargs
    ):
        """Enter a layer context, indenting all future log messages.

        Args:
            msg (str): message
            label (str, optional): The label added before the message (if specified in format string). Defaults to None.
            owner (Str, optional): Owner of the layer, which is added to the stack. Defaults to None.
            prefix (str, optional): Prefix added before the message. Defaults to "type".
        """
        kwargs.update({'stacklevel': kwargs.get('stacklevel', 1)+1})
        self.header(msg, label, prefix, *args, **kwargs)
        self.stack.append(owner)
        yield self
        self.stack.pop()
        self.header(msg, f"end {label}", prefix, *args, **kwargs)

    def makeRecord(self, *args, **kwargs):
        rv = super().makeRecord(*args, **kwargs)
        trace: Trace = getattr(rv, 'trace', None)
        if trace:
            rv.exc_text = ''.join(trace.format())
            rv.exc_info = None
        return rv
    
    def _log(self, level, msg, args, exc_info = None, extra=None, stack_info: bool=False, stacklevel: int=1) -> None:
        return super()._log(level, msg, args, exc_info, extra, stack_info, stacklevel+2)



logger: Logger = backend.getLogger(LOGGER_NAME)
logger.__class__ = Logger
logger.propagate = False
logger.stack = []


def info(
    msg: str, label: str = None, color: str = None, use_rich_highlighter: bool = False
):
    """Log an info message.

    Args:
        msg (str): message
        label (str, optional): The label added before the message (if specified in format string). Defaults to None.
        color (str, optional): The color of the text. Defaults to None.
        use_rich_highlighter (bool, optional): Use rich highlighting. Defaults to False.
    """
    if len(logger.handlers) == 0:
        default_config()
    return logger.info(msg, label, color, use_rich_highlighter, stacklevel=2)


def debug(msg: str, label: str = None):        
    """Log a debug message.

    Args:
        msg (str): message
        label (str, optional): The label added before the message (if specified in format string). Defaults to None.
    """
    if len(logger.handlers) == 0:
        default_config()
    return logger.debug(msg, label, stacklevel=2)


def warning(msg: str, label: str = None, exc_info=False):
    """Log a warning message

    Args:
        msg (str): message
        label (str, optional): The label added before the message (if specified in format string). Defaults to None.
        exc_info (bool, optional): Add exception info. Defaults to False.
    """
    if len(logger.handlers) == 0:
        default_config()
    return logger.warning(msg, label, exc_info, stacklevel=2)


def error(msg: str, label: str = None, exc_info=False, stacklevel=1):
    """Log an error message

    Args:
        msg (str): message
        label (str, optional): The label added before the message (if specified in format string). Defaults to None.
        exc_info (bool, optional): Add exception info. Defaults to False.
    """
    if len(logger.handlers) == 0:
        default_config()
    return logger.error(msg, label, exc_info, stacklevel=2)


def exception(msg: str, label: str = None, exc_info=True):
    """Log an exception

    Args:
        msg (str): message
        label (str, optional): The label added before the message (if specified in format string). Defaults to None.
        exc_info (bool, optional): Add exception info. Defaults to True.
    """
    if len(logger.handlers) == 0:
        default_config()
    return logger.exception(msg, label, exc_info, stacklevel=2)


def emphasize(msg: str, label: str = None):
    """Log an emphasized info message.

    Args:
        msg (str): message
        label (str, optional): The label added before the message (if specified in format string). Defaults to None.
    """
    if len(logger.handlers) == 0:
        default_config()
    return logger.emphasize(msg, label, stacklevel=2)


def io(msg: str, label: str = None):
    """Log an info message associated with io.

    Args:
        msg (str): message
        label (str, optional): The label added before the message (if specified in format string). Defaults to None.
    """
    if len(logger.handlers) == 0:
        default_config()
    return logger.io(msg, label, stacklevel=2)


def layer(msg: str, label: str = None, prefix: str = "type", owner: str = None):
    """Enter a layer context, indenting all future log messages.

    Args:
        msg (str): message
        label (str, optional): The label added before the message (if specified in format string). Defaults to None.
        owner (Str, optional): Owner of the layer, which is added to the stack. Defaults to None.
        prefix (str, optional): Prefix added before the message. Defaults to "type".
    """
    if len(logger.handlers) == 0:
        default_config()
    return logger.layer(msg, label, prefix, owner, stacklevel=2)
        

@contextmanager
def logger_context(level: bool = None):
    """Enter a logger context with a temporary log level.
    
    Example:

    >>> with logger_context(level=INFO):
    ...     info('test', 'label')
    """
    if level is None:
        yield logger
        return

    _level = logger.level
    logger.setLevel({True: INFO, False: WARNING}.get(level, level))
    yield logger
    logger.setLevel(_level)


from dataclasses import dataclass, field
from typing import List, Type, Any
from types import TracebackType
import traceback
import re


@dataclass
class Frame:
    filename: str
    lineno: int
    name: str
    line: str = ""

    def __post_init__(self):
        if isinstance(self.lineno, str):
            self.lineno = int(self.lineno)

    def __str__(self):
        return f"File {self.filename}, line {self.lineno}, in {self.name}: {self.line}"

    def __serialize__(self):
        return str(self)

    @classmethod
    def __deserialize__(cls, ser):
        pattern = r"File (?P<filename>.*), line (?P<lineno>[0-9]*), in (?P<name>.*): (?P<line>)"
        res, *_ = re.findall(pattern, ser)
        return cls(*res)
    
    def summary(self):
        return traceback.FrameSummary(self.filename, self.lineno, self.name)


@dataclass
class Stack:
    exc_type: str
    exc_value: str
    is_cause: bool = False
    frames: List[Frame] = field(default_factory=list)

    def __serialize__(self):
        res = {"exc": f"{self.exc_type}({self.exc_value})"}
        if self.is_cause:
            res["is_cause"] = True
        res["frames"] = [frame.__serialize__() for frame in self.frames]
        return res
    
    def __str__(self):
        return self.format_exception()

    @classmethod
    def __deserialize__(cls, ser: dict):
        exc = ser.pop("exc")
        pattern = r"(?P<filename>.*)\((?P<name>.*)\)"
        (ser["exc_type"], ser["exc_value"]), *_ = re.findall(pattern, exc)
        res = cls(**ser)
        res.frames = [Frame.__deserialize__(frame) for frame in res.frames]
        return res

    def format_exception(self):
        yield f'{self.exc_type}: {self.exc_value}'
    
    def format(self, *, single: bool = False):
        if single:
            yield '\n'
        elif not self.is_cause:
            yield '\n\nThe above exception was the direct cause of the following exception:\n\n'
            
        if len(self.frames) > 0:
            yield 'Traceback (most recent call last):\n'
            yield from traceback.StackSummary.from_list(
                [frame.summary() for frame in self.frames]
            ).format()

        yield from self.format_exception()


@dataclass
class Trace:
    stacks: List[Stack] = field(default_factory=list)

    def __serialize__(self):
        return [stack.__serialize__() for stack in self.stacks]

    @classmethod
    def __deserialize__(cls, ser: list):
        return cls([Stack.__deserialize__(stack) for stack in ser])

    @classmethod
    def from_exception(
        cls,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_tb: TracebackType,
        *,
        ignore: int = 0
    ):
        def safe_str(_object: Any) -> str:
            try:
                return str(_object)
            except Exception:
                return "<exception str() failed>"

        stacks: List[Stack] = []
        is_cause = False

        while True:
            summary = traceback.extract_tb(exc_tb)
            frames = [Frame(f.filename, f.lineno, f.name, f.line) for f in summary]
            stacks.append(
                Stack(
                    safe_str(exc_type.__name__), safe_str(exc_value), is_cause, frames
                )
            )

            cause = getattr(exc_value, "__cause__", None)
            if cause and cause.__traceback__:
                exc_type = cause.__class__
                exc_value = cause
                exc_tb = cause.__traceback__
                is_cause = True
                continue

            cause = exc_value.__context__
            if (
                cause
                and cause.__traceback__
                and not getattr(exc_value, "__suppress_context__", False)
            ):
                exc_type = cause.__class__
                exc_value = cause
                exc_tb = cause.__traceback__
                is_cause = False
                continue
            # No cover, code is reached but coverage doesn't recognize it.
            break  # pragma: no cover

        res = cls(stacks)
        if ignore > 0:
            res.stacks[0].frames = res.stacks[0].frames[ignore:]
        return res
    
    def format(self):
        for stack in reversed(self.stacks):
            yield from stack.format(single=len(self.stacks) == 1)


@dataclass(repr=False)
class BaseInvalid:
    """Invalid object encountered by ``dman``."""
    type: str   #: The name of the type.
    info: str   #: Description of the problem.

    def __repr__(self):
        return f'{self.header} ({self.info})'

    def format(self):
        """Generator producing the description of the invalid object."""
        yield self.header + '\n'
        yield ' '*4 + self.info
    
    @property
    def header(self):
        """Header of the description."""
        return f'{self.__class__.__name__}: {self.type}'

    def __str__(self):
        return ''.join(self.format())

    def __serialize__(self):
        res = {'type': self.type}
        if len(self.info) > 0:
            res['info'] = self.info
        return res

    @classmethod
    def __deserialize__(cls, ser: dict):
        return cls(ser.get('type'), ser.get('info', ''))


@dataclass(repr=False)
class ExcInvalid(BaseInvalid):
    """Invalid object created through some exception."""
    trace: Trace    #: The traceback of the exception.

    @classmethod
    def from_exception(
        cls,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_tb: TracebackType,
        *,
        ignore: int = 0,
        **kwargs,
    ):
        """Create an invalid object from an exception.

        Args:
            exc_type (Type[BaseException]): The type of the exception
            exc_value (BaseException): The value of the exception
            exc_tb (TracebackType): The traceback
            ignore (int, optional): Layers of the traceback that should be omitted. Defaults to 0.
        """
        trace = Trace.from_exception(exc_type, exc_value, exc_tb, ignore=ignore)
        return cls(**kwargs, trace=trace)

    def format(self):
        yield from super().format()
        yield from self.trace.format()
    
    def __serialize__(self):
        return dict(
            **super().__serialize__(), 
            trace=self.trace.__serialize__()
        )
    
    @classmethod
    def __deserialize__(cls, ser):
        return cls(
            ser.get('type'), 
            ser.get('info', ''), 
            Trace.__deserialize__(ser.get('trace'))
        )



"""
Toolbox for experimental data management in Python.
"""

__version__ = '1.0'

from dman.core.serializables import serializable, serialize, deserialize, BaseContext, is_serializable
from dman.core.serializables import ser_type2str, ser_str2type, register_serializable, register_instance
from dman.core.serializables import isvalid, ValidationError, SerializationError
from dman.core.storables import FileTarget
from dman.core.storables import storable, write, read, register_storable
from dman.core.path import mount, target, get_root_path, AUTO
from dman.core import log

from dman.model.record import record, remove, Context
from dman.model.modelclasses import recordfield, mdict, smdict, mlist, smlist, mruns, smruns, serializefield, modelclass
from dman.model.modelclasses import smlist_factory, smdict_factory, smruns_factory, mruns_factory, mdict_factory, mlist_factory
from dman.model.modelclasses import record_fields, unused_fields, register_preset
from dman.model.repository import track, save, load, store, clean
from dman.model.repository import uninterrupted, context

from dman.utils import sjson
from dman.utils.smartdataclasses import idataclass, configclass, optionfield, is_configclass

from dman.config import params

try:
    import dman.numeric as numeric
    from dman.numeric import barray, sarray, carray
    _numeric_available = True
except ImportError as e:
    _numeric_available = False
    barray, sarray, carray = None, None, None

try:
    import dman.tui as tui
    _tui_available = True
except ImportError as e:
    _tui_available = False


from dataclasses import field, dataclass




import rich

from typing import Dict, Iterable, Sequence, Tuple
import os
import pathlib
from collections import OrderedDict

from rich.style import Style
from rich.console import JustifyMethod, Console, Group
from rich.table import Table
from rich.panel import Panel
from rich.columns import Columns
from rich import box
from rich.progress import Progress, track
from rich.live import Live
from rich.tree import Tree
from rich.markup import escape
from rich import inspect
from rich.filesize import decimal
from rich.markup import escape
from rich.text import Text
from rich.json import JSON
from rich.syntax import Syntax
from rich.pretty import pprint

from dman.core.path import get_root_path, normalize_path
from dman.core.serializables import BaseContext, serialize
from dman import sjson

_print = print

from rich import print


def print_serializable(obj, context: BaseContext = None, content_only: bool = False):
    print_serialized(serialize(obj, context=context, content_only=content_only))


def print_serialized(ser):
    print_json(sjson.dumps(ser, indent=4))


def print_json(json: str):
    print(JSON(json))



def progress(it: Iterable, description: str = None, show_state: bool = True, total = None):
    if total is None and hasattr(it, "__len__"):
        total = len(it)
    if description is None:
        description = f'Iterating "{it.__class__.__name__}"'
    if total and show_state:
        description = description + ' [{i}/{t}]'
        def gen(i, t):
            return description.format(i=i, t=t)
    else:
        total = 100.0
        def gen(i, t):
            return description
        
    with Progress() as progress:
        task = progress.add_task(gen(0, total), total=total)
        for i, o in enumerate(it):
            progress.update(task, advance=1, description=gen(i+1, total), refresh=True)
            yield o


class StackGenerator:
    def __init__(self, state: Tuple[int] = None):
        self.progress = None
        self._state = [] if state is None else list(state)
        self.registered: Dict[StackLayer, int] = dict()
        self.index: Dict[StackLayer, int] = OrderedDict()

    @property
    def state(self):
        return list(self.index.values())

    def print(self, msg: str):
        self.progress.print(msg)

    def __call__(
        self,
        it: Iterable,
        *,
        total: int = None,
        keep: bool = None,
        description: str = None,
        post: str = ' ...',
        show_state: bool = True,
        log: dict = None,
    ):
        if keep is None:
            keep = len(self.registered) == 0
        if total is None and hasattr(it, "__len__"):
            total = len(it)
        if log is None:
            log = dict()
        logstr = ''
        for k in log:
            logstr += (str(k) + '={' + str(k) +'}, ')
        if len(logstr) > 0:
            logstr = ' | ' + logstr[:-2] + ' |' 
        if description is None:
            description = f'Iterating "{it.__class__.__name__}"'
        if total and show_state:
            description = description + ' [{state}/{total}]'
            log.update({'state': 1, 'total': total})
        description += logstr + post

        state = 0 if len(self._state) == 0 else self._state.pop(0)
        layer = StackLayer(it, self, state, keep, description, log)
        layer._update(total=total)
        self.registered[layer] = self.add_task(layer.description, total=total)
        self.index[layer] = 0
        return layer

    def range(self, *args, description: str = None, **kwargs):
        if description is None: description = 'Range'
        it = range(*args)
        return self.__call__(it, description=description, **kwargs)

    def add_task(self, name, total: int = None):
        if self.progress is None:
            return None
        return self.progress.add_task(name, total=total)

    def remove_task(self, task: "StackLayer"):
        if self.progress is None:
            return
        self.progress.remove_task(self.registered[task])
        del self.index[task]

    def update(self, task: "StackLayer", completed: int, description: str = None):
        if self.progress is None:
            return
        self.index[task] += 1
        self.progress.update(self.registered[task], completed=completed, description=description, refresh=True)
    
    def start(self):
        self.progress = Progress()
        self.progress.__enter__()

    def stop(self):
        self.progress.stop()
        self.progress = None
        for r in self.registered:
            r.progress = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *_):
        self.stop()


class StackLayer:
    def __init__(
        self,
        it: Iterable,
        parent: StackGenerator,
        state: int = 0,
        keep: bool = False,
        description: str = None,
        log: dict = None
    ):        
        self._description = description
        self.log = dict() if log is None else log
        self.parent = parent
        self.skip = state
        self.keep = keep
        self.state = 0
        self.it = it

    @property
    def description(self):
        return str.format(self._description, **self.log)

    def __iter__(self):
        for x in self.it:
            if self.state >= self.skip:
                yield x
            self.state += 1
            self.update(state=min(self.log.get('total', 0), self.state+1))
        if not self.keep:
            self.parent.remove_task(self)
    
    def _update(self, **kwargs):
        for k, v in kwargs.items():
            if k in self.log:
                self.log[k] = v

    def update(self, **kwargs):
        self._update(**kwargs)
        self.parent.update(self, completed=self.state, description=self.description)


def stack(state: Tuple[int] = None):
    return StackGenerator(state)


def walk_file(path: pathlib.Path):
    text_chars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F})
    with open(path, "rb") as f:
        is_binary_string = bool(f.read(1024).translate(None, text_chars))

    if is_binary_string:
        return None
    with open(path, "r") as f:
        content = f.read()
        if path.suffix == ".json":
            return Panel(JSON(content), box=box.HORIZONTALS)
        # if path.suffix == ".py":
        #     return Panel(Syntax(content), "python")
        return Panel(Text(content), box=box.HORIZONTALS)


def walk_directory(
    directory: pathlib.Path,
    *,
    show_content: bool = False,
    tree: Tree = None,
    normalize: bool = True,
    show_hidden: bool = False,
    console: Console = None,
) -> None:
    """Print the contents of a directory

    :param directory: directory to print
    :param show_content: show content of text files, defaults to False. If True 
        all non-binary files are shown. If a list of strings then only files with that 
        extension are shown. For example show_content=['.json'] shows json files only.
    :param tree: add content to tree instead of printing, defaults to None
    :param normalize: normalize paths with respect to dman root directory, defaults to True
    :param show_hidden: show hidden files, defaults to False
    :param console: console for printing
    """

    # based on https://github.com/Textualize/rich/blob/master/examples/tree.py
    is_root = tree is None
    if is_root:
        name = directory
        if normalize:
            name = normalize_path(name)
        tree = Tree(
            f":open_file_folder: [link file://{directory}]{name}",
            guide_style="bold bright_blue",
        )

    # sort dirs first then by filename
    paths = sorted(
        pathlib.Path(directory).iterdir(),
        key=lambda path: (path.is_file(), path.name.lower()),
    )

    for path in paths:
        # remove hidden files
        if not show_hidden and path.name.startswith("."):
            continue
        if path.is_dir():
            style = "dim" if path.name.startswith("__") else ""
            branch = tree.add(
                f"[bold magenta]:open_file_folder: [link file://{path}]{escape(path.name)}",
                style=style,
                guide_style=style,
            )
            walk_directory(
                path, tree=branch, show_content=show_content, show_hidden=show_hidden
            )
        else:
            text_filename = Text(path.name, "green")
            text_filename.highlight_regex(r"\..*$", "green")
            text_filename.stylize(f"link file://{path}")
            file_size = path.stat().st_size
            text_filename.append(f" ({decimal(file_size)})", "blue")
            icon = "🐍 " if path.suffix == ".py" else "📄 "

            res = Text(icon) + text_filename
            if (
                (
                    isinstance(show_content, Sequence)
                    and os.path.splitext(path.name)[-1] in show_content
                ) or (
                    isinstance(show_content, bool)
                    and show_content == True)
            ):
                content = walk_file(path)
                if content is not None:
                    res = Group(res, content)
            tree.add(res)

    if is_root:
        if console is None:
            console = Console(width=80)
        console.print(tree)



import json
from typing import Callable, Type


def _default(o):
    convert = __translate_atomic.get(
        type(o), 
        lambda o: f"<un-serializable: {type(o).__qualname__}>"
    )
    return convert(o)


def dumps(obj, *args, **kwargs):
    return json.dumps(obj, *args, default=_default, **kwargs)


def dump(obj, *args, **kwargs):
    return json.dump(obj, *args, default=_default, **kwargs)


def load(fp, *, cls=None, object_hook=None, parse_float=None,
        parse_int=None, parse_constant=None, object_pairs_hook=None, **kw):
    return json.load(
        fp, cls=cls, object_hook=object_hook, 
        parse_float=parse_float, parse_int=parse_int, 
        parse_constant=parse_constant, 
        object_pairs_hook=object_pairs_hook, **kw
    )


def loads(s, *, cls=None, object_hook=None, parse_float=None,
          parse_int=None, parse_constant=None, object_pairs_hook=None, **kw):
    return json.loads(
        s, cls=cls, object_hook=object_hook,
        parse_float=parse_float, parse_int=parse_int,
        parse_constant=parse_constant,
        object_pairs_hook=object_pairs_hook, **kw
    )


atomic_types = [
    str, int, float, bool, type(None)
]

__translate_atomic = {}


def atomic_type(obj):
    return isinstance(obj, tuple(atomic_types))


def register_atomic_alias(obj: Type, convert: Callable):
    atomic_types.append(obj)
    __translate_atomic[obj] = convert



def prompt_user_bool(question, default=None):
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        print(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' " "(or 'y' or 'n').")


def prompt_user(question, default: str = None):
    if default is None:
        print(question, end=" ")
        return input().lower()
    else:
        print(question + f" [{default}]:", end=" ")
        choice = input().lower()
        return choice if len(choice) > 0 else default


"""
Defines records, which are used to make ``storable`` objects ``serializable``.
"""

import os
import sys
import uuid

from dataclasses import asdict, dataclass, is_dataclass
from typing import Any, Tuple
from contextlib import suppress

from dman.core import log
from dman.core.serializables import (
    SER_CONTENT,
    SER_TYPE,
    is_serializable,
    serializable,
    BaseContext,
    serialize,
    SerializationError,
    isvalid,
    ExcUnserializable,
    Unserializable,
)
from dman.core.storables import (
    STO_TYPE,
    is_storable,
    read,
    sto_type2str,
    get_storable_name,
    write,
)
from dman.core.path import (
    Target,
    AUTO,
    Mount,
    mount,
    UserQuitException,
    MountException,
    config,
)


REMOVE = "__remove__"
EXTENSION = "__ext__"


def is_removable(obj):
    """Check if an object is removable."""
    return hasattr(obj, REMOVE)


@serializable(name="__no_file")
class NoFile(Unserializable):
    """Returned when no file was found in the record target."""
    def __repr__(self):
        return f"NoFile: {self.type}"


@serializable(name="__un_writable")
class UnWritable(Unserializable):
    """Returned when no data could be written to the record target file."""
    def __repr__(self):
        return f"UnWritable: {self.type}"


@serializable(name="__exc_un_writable")
class ExcUnWritable(ExcUnserializable):
    """Returned when no data could be written to the 
        record target file due to an exception."""
    def __repr__(self):
        return f"ExcUnWritable: {self.type}"


@serializable(name="__un_readable")
@dataclass
class UnReadable(Unserializable):
    """Returned when the record target file could not be read."""
    target: str

    def __repr__(self):
        return f"UnReadable: {self.type}"

    def __serialize__(self):
        res = super().__serialize__()
        res["target"] = self.target
        return res

    @classmethod
    def __deserialize__(cls, ser: dict, context):
        target = ser.pop("target")
        res: Unserializable = Unserializable.__deserialize__(ser, context)
        return cls(type=res.type, info=res.info, target=target)


@serializable(name="__exc_un_readable")
@dataclass
class ExcUnReadable(ExcUnserializable):
    """Returned when the record target file 
        could not be read due to an exception."""
    target: str

    def __serialize__(self):
        res = super().__serialize__()
        res["target"] = self.target
        return res

    @classmethod
    def __deserialize__(cls, ser: dict, context):
        target = ser.pop("target")
        res: ExcUnserializable = ExcUnserializable.__deserialize__(ser, context)
        return cls(res.type, res.info, res.trace, target=target)


class StoreError(SerializationError):
    """Raised when storing failed in such a way that 
        serialization should be cancelled. Usually caused by 
        the quit option was passed when a file is overwritten."""    
    ...


class Context(BaseContext):
    """Serialization context that keeps track of the current folder."""
    def __init__(
        self,
        mnt: Mount,
        subdir: os.PathLike = "",
    ):
        """Initialize a context with a mount point and a subdirectory relative to it."""
        super().__init__()
        self.mnt = mnt
        self.subdir = subdir

    @property
    def directory(self):
        """Absolute directory that the context controls."""
        return self.mnt.abspath(self.subdir, validate=False)

    def __repr__(self):
        return f"Context({self.directory})"

    @classmethod
    def from_directory(cls, directory: str, gitignore: bool = True):
        """Create a context based on a directory."""
        return cls(Mount(directory, gitignore=gitignore))

    @classmethod
    def mount(
        cls,
        key: str,
        *,
        subdir: os.PathLike = "",
        cluster: bool = True,
        generator: str = None,
        base: os.PathLike = None,
        gitignore: bool = True,
    ):
        """Get a context from a mount point.
        The path of the file is determined as described below.

            If the files are clustered then the path is ``<base>/<generator>/<subdir>/<key>/<key>.<ext>``
            If cluster is set to False then the path is ``<base>/<generator>/<subdir>/<key>.<ext>``

            When base is not provided then it is set to .dman if
            it does not exist an exception is raised.

            When generator is not provided it will automatically be set based on
            the location of the script relative to the .dman folder
            (again raising an exception if it is not found). For example
            if the script is located in ``<project-root>/examples/folder/script.py``
            and .dman is located in ``<project-root>/.dman``.
            Then generator is set to cache/examples:folder:script (i.e.
            the / is replaced by : in the output).

        Args:
            key (str):  Key for the file.
            subdir (os.PathLike, optional): Specifies optional subdirectory in generator folder. Defaults to "".
            cluster (bool, optional): A subfolder ``key`` is automatically created when set to True. Defaults to True.
            generator (str, optional): Specifies the generator that created the file. Defaults to script label.
            base (os.PathLike, optional): Specifies the root folder. Defaults to ".dman".
            gitignore (bool, optional): Specifies whether files added to this mount point should be ignored.
        """
        return cls(
            mount(
                key,
                subdir=subdir,
                cluster=cluster,
                generator=generator,
                base=base,
                gitignore=gitignore,
            )
        )

    def serialize_list(self, ser: list):
        """Serialize a list.
            Automatically turns the list into an :class:`dman.mlist` if
            it contains a storable object.
        """
        res = []
        is_model = False
        for itm in ser:
            if is_storable(itm):
                itm = record(itm)
                is_model = True
            res.append(self.serialize(itm))
        if is_model:
            res = {SER_TYPE: "_ser__mlist", SER_CONTENT: {"store": res}}
        return res

    def serialize_dict(self, ser: dict):
        """Serialize a dict.
            Automatically turns the dict into an :class:`dman.mdict` if
            it contains a storable object.
        """
        res = {}
        is_model = False
        for k, v in ser.items():
            if is_storable(v):
                v = record(v)
                is_model = True
            k = self.serialize(k)
            res[k] = self.serialize(v)
        if is_model:
            res = {SER_TYPE: "_ser__mdict", SER_CONTENT: {"store": res}}
        return res

    def absolute(self, target: os.PathLike):
        """Get the path of a target relative to the mount point."""
        if not isinstance(target, Target):
            target = Target.from_path(target)
        return target.update(subdir=os.path.join(self.subdir, target.subdir))

    def write(self, target: os.PathLike, storable):
        """Write a storable to a target."""
        try:
            target = Target.from_path(target)
            local, _target = self.prepare(target)
            path = self.mnt.abspath(local.absolute(_target))
            return target.update(name=_target.name), write(storable, path, local)
        except SerializationError:
            raise
        except UserQuitException as e:
            raise StoreError(*e.args)
        except Exception:
            res = ExcUnWritable.from_exception(
                *sys.exc_info(),
                type=get_storable_name(storable),
                info="Exception encountered while writing.",
                ignore=4,  # TODO verify
            )
            self.process_invalid("An error occurred while writing.", res)
            return target.update(name=_target.name), res

    def read(self, target: os.PathLike, expected):
        """Read a storable of some expected type from a target."""
        try:
            local, target = self.prepare(target, choice="_ignore")
            path = self.mnt.abspath(local.absolute(target))
            return read(expected, path, local)
        except FileNotFoundError:
            if not isinstance(expected, str):
                expected = getattr(expected, STO_TYPE, None)
            res = NoFile(type=expected, info=f"Missing Target: {target}.")
            self.process_invalid("Could not find specified file.", res)
            return res
        except SerializationError:
            raise
        except Exception:
            res = ExcUnReadable.from_exception(
                *sys.exc_info(),
                type=get_storable_name(expected),
                info="Exception encountered while reading.",
                target=target,
                ignore=4,  # TODO verify
            )
            self.process_invalid("An error occurred while reading.", res)
            return res

    def remove(self, target: os.PathLike = None, *, obj=None):
        """Remove object or file, stored at specified target."""
        if target is None:
            local = self
        else:
            # Get target with respect to mount point
            target = self.absolute(target)

            # Remove file if it exists.
            try:
                self.mnt.remove(target)
            except MountException:
                log.warning(
                    f'Tried to remove file outside of mount point: "{target}".',
                    "context",
                )

            # Localize the target
            local, target = self.join(target.subdir), Target(name=target.name)

        # Remove files created by the object
        tp = type(obj)
        if is_removable(obj):
            with log.layer(tp.__name__, "remove"):
                inner_remove = getattr(obj, REMOVE, None)
                if inner_remove is not None:
                    inner_remove(local)
                return

        if is_dataclass(obj):
            obj = asdict(obj)

        if isinstance(obj, (tuple, list)):
            with log.layer(tp.__name__, "remove"):
                for x in obj[:]:
                    local.remove(obj=x)
        elif isinstance(obj, dict):
            with log.layer(tp.__name__, "remove"):
                for k in obj.keys():
                    local.remove(obj=obj[k])

    def join(self, subdir: os.PathLike):
        """Create a context relative to this one based on the subdirectory."""
        if subdir == "":
            return self
        return self.__class__(mnt=self.mnt, subdir=subdir)
    
    def open(self, target: os.PathLike, *args, **kwargs):
        """Open a file, registered by this context.
        
            The signature is identical to the standard ``open`` command.
        """
        return self.mnt.open(self.absolute(target), *args, **kwargs)

    def prepare(
        self, target: os.PathLike, *, choice: str = None
    ) -> Tuple["Context", Target]:
        """Prepare a target for writing a storable to."""
        target = self.mnt.prepare(self.absolute(target), choice=choice)
        return self.join(target.subdir), Target(name=target.name)

    def __enter__(self):
        self.mnt.__enter__()
        return self

    def __exit__(self, *_):
        self.mnt.__exit__(*_)

    def close(self):
        """Close this context point. 
        
            Empty subdirectories are deleted and a gitignore is created if 
            requested on creation.
        """
        self.mnt.close()


def is_unloaded(obj):
    """Check if an object is unloaded."""
    return isinstance(obj, Unloaded)


def is_undefined(obj):
    """Check if an object is undefined."""
    return isinstance(obj, Undefined)


@dataclass
class Unloaded:
    """Unloaded storable object."""
    type: str               #: Storable type
    target: os.PathLike     #: Target relative to the context
    base: os.PathLike       #: The directory of the context
    context: Context        #: The context used when the record was deserialized.

    @property
    def path(self):
        """Path where the content of this storable is located."""
        return os.path.join(self.base, self.target)

    def __load__(self):
        return self.context.read(self.target, self.type)

    def __repr__(self) -> str:
        return f"UL[{self.type}]"


@dataclass
class Undefined:
    """Undefined storable object."""
    type: str   #: Storable type

    def __repr__(self) -> str:
        return f"ERR[{self.type}]"


@serializable(name="_record_exceptions")
@dataclass
class _RecordExceptions:
    write: Any = None
    read: Any = None

    def __serialize__(self):
        res = {}
        if self.write is not None:
            res["write"] = serialize(self.write)
        if self.read is not None:
            res["read"] = serialize(self.read)
        return res

    @classmethod
    def __deserialize__(cls, ser: dict):
        return cls(ser.get("write", None), ser.get("read", None))

    def empty(self):
        return self.write is None and self.read is None


@serializable(name="_ser__record")
class Record:
    """Wraps a storable in such a way to make it serializable. 
    
        Keeps track of the file the storable is written to and 
        by default only loads the content when it is requested.

        When no target is given a unique file name is determined using 
        ``uuid4``. The suffix of the file is determined automatically 
        based on the type.
    """

    def __init__(self, content: Any, target: Target = None, preload: bool = False):
        """Create a new record

        Args:
            content (Any): The storable contained within.
            target (Target, optional): The target to which to write.
            preload (bool, optional): Load the storable upon deserialization. Defaults to False.
        """
        self._content = content
        self._target = Target() if target is None else target

        self.preload = preload
        self.exceptions = _RecordExceptions()

    def exists(self):
        """Check if the file associated with this record exists."""
        return not isinstance(self.exceptions.read, NoFile)

    def isvalid(self, *, load: bool = False):
        """Check whether the content was loaded successfully."""
        if load:
            self.load()
        return self.exceptions.read is None

    @property
    def target(self):
        """The target path where content is written."""
        if self._target.is_complete():
            return self._target

        base = Target(stem=f"{uuid.uuid4()}", subdir=f"{uuid.uuid4()}")
        if is_serializable(self._content) or is_dataclass(self._content):
            base = base.update(suffix=".json")
        else:
            base = base.update(suffix=config.default_suffix)
        request = Target(suffix=getattr(self._content, EXTENSION, AUTO))
        self._target = base.merge(request, self._target)
        return self._target

    @property
    def sto_type(self):
        """The storable type string of the content."""
        if is_unloaded(self._content) or is_undefined(self._content):
            return self._content.type
        return sto_type2str(self._content)

    @property
    def content(self):
        """Get the content of this method, loading it if not done before."""
        return self.load()

    @content.setter
    def content(self, value: Any):
        """Assign a storable value to the content of this record."""
        if not is_storable(value):
            raise ValueError("Expected storable type.")
        self._content = value

    def __repr__(self) -> str:
        content = self._content
        preload_str = ""
        if self.preload:
            preload_str = ", preload"
        if content is None:
            return f"Record(None, target={self.target}{preload_str})"
        if is_unloaded(content) or is_undefined(content):
            return f"Record({content}, target={self.target}{preload_str})"
        return f"Record({sto_type2str(content)}, target={self.target}{preload_str})"

    def load(self):
        """Load the content of this record from the file."""
        if not is_unloaded(self._content):
            return self._content

        ul: Unloaded = self._content
        with log.layer("loading content", "record", owner="record"):
            content = ul.__load__()
            if isvalid(content):
                self.exceptions.read = None
            else:
                self.exceptions.read = content
                return content
            self._content = content
        return self._content

    def store(self, context: BaseContext):
        """Store the content of this record."""
        if isinstance(context, Context):
            if is_unloaded(self._content) and self._content.base != context.directory:
                self.load()  # the target was moved
            elif self.exceptions.write is not None:
                self.load()  # the previous store failed
            elif is_unloaded(self._content):
                return self._content.target  # no load needed

            # execute store
            target, self.exceptions.write = context.write(self.target, self._content)
            return target
        else:
            self.exceptions.write = UnWritable(
                type=self.sto_type,
                info=f"Invalid context passed to Record({self.sto_type}, target={self.target}).",
            )
            context.process_invalid(
                "Exception encountered during write.", self.exceptions.write
            )
            return self.target

    def __serialize__(self, context: BaseContext):
        sto_type = self.sto_type
        target: Target = self.store(context)
        self._target = target
        res = {
            "target": str(target) if target.is_complete() else tuple(target),
            "sto_type": sto_type,
        }
        if self.preload:
            res["preload"] = self.preload
        if not self.exceptions.empty():
            res["exceptions"] = serialize(self.exceptions, context, content_only=True)

        return res

    @classmethod
    def __deserialize__(cls, serialized: dict, context: BaseContext):
        target = serialized.get("target", None)
        if isinstance(target, str):
            target = Target.from_path(target)
        if isinstance(target, list):
            target = Target(*target)
        sto_type = serialized.get("sto_type")
        preload = serialized.get("preload", False)

        # load previous exceptions
        exceptions = serialized.get("exceptions", None)
        if exceptions is None:
            exceptions = _RecordExceptions()
        else:
            exceptions = _RecordExceptions.__deserialize__(exceptions)

        # try to load the contents of the record
        if isinstance(context, Context) and isinstance(target, Target):
            content = Unloaded(sto_type, target, context.directory, context)
        elif not isinstance(context, Context):
            exceptions.read = UnReadable(
                type=sto_type,
                info=f"Invalid context passed to Record({sto_type}, target={target}).",
                target=target,
            )
            context.process_invalid(
                "Exception encountered during read.", exceptions.read
            )
            content = Undefined(sto_type)
        else:
            exceptions.read = UnReadable(
                type=sto_type,
                info=f"Invalid target recovered Record({sto_type}, target={target}).",
                target=target,
            )
            context.process_invalid(
                "Exception encountered during read.", exceptions.read
            )
            content = Undefined(sto_type)

        out = cls(content=content, target=target, preload=preload)
        if preload:
            out.load()
        out.exceptions = exceptions
        return out

    def __remove__(self, context: BaseContext):
        if isinstance(context, Context):
            if isvalid(self.content):
                return context.remove(obj=self.content, target=self.target)
            log.warning(
                "Loaded content is invalid. Could not continue removing.", "record"
            )
        else:
            log.warning("Tried removing with invalid context type.", "record")


def record(
    content: Any,
    /,
    *,
    stem: str = AUTO,
    suffix: str = AUTO,
    name: str = AUTO,
    subdir: os.PathLike = "",
    preload: str = False,
):
    """
    Wrap a storable object in a serializable record.
        The target path is specified as:

        * ``./subdir/stem+suffix`` or
        * ``./subdir/name``.

    :param content:         The storable object.
    :param str stem:        The stem of a file.
    :param str suffix:      The suffix or extension of the file (e.g. ``'.json'``).
    :param str name:        The full name of the file.
    :param str subdir:      The subdirectory in which to store te file.
    :param bool preload:    When ``True`` the file will be loaded during deserialization.

    :raises ValueError:     if a name and a stem and/or suffix are specified.
    """
    return Record(
        content,
        Target(stem=stem, suffix=suffix, name=name, subdir=subdir),
        preload,
    )


def remove(obj, context: Context = None):
    """Remove all files created by an object."""
    if not isinstance(context, Context):
        return
    context.remove(obj=obj)



import pickle
import dman

import shutil
from typing import Union, List
import subprocess
import re
import textwrap

import os
import math



try:
    import matplotlib
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages
except ImportError as e:
    raise ImportError('Plotting tools require matplotlib.') from e


@dman.storable(name='__plt_eps')
class PrintFigure:
    __ext__ = '.eps'

    def __init__(self, fig: plt.Figure, ext: str = '.eps', **kwargs):
        self.__ext__ = ext
        self.fig = fig
        self.kwargs = kwargs

    def __write__(self, path: str):
        self.fig.savefig(path, **self.kwargs)

    @classmethod
    def __read__(self):
        raise RuntimeError('Cannot read PrintFigure objects.')


class SVGLogFormatter(matplotlib.ticker.LogFormatterMathtext):
    def _non_decade_format(self, sign_string, base, fx, usetex):
        """Return string for non-decade locations."""
        return r'\$%s%s^{%.2f}\$' % (sign_string, base, fx)

    def __call__(self, x, pos=None):
        # docstring inherited
        min_exp = plt.rcParams['axes.formatter.min_exponent']

        if x == 0:  # Symlog
            return r'\$0\$'

        sign_string = '-' if x < 0 else ''
        x = abs(x)
        b = self._base

        # only label the decades
        fx = math.log(x) / math.log(b)
        is_x_decade = matplotlib.ticker.is_close_to_int(fx)
        exponent = round(fx) if is_x_decade else np.floor(fx)
        coeff = round(b ** (fx - exponent))
        if is_x_decade:
            fx = round(fx)

        if self.labelOnlyBase and not is_x_decade:
            return ''
        if self._sublabels is not None and coeff not in self._sublabels:
            return ''

        # use string formatting of the base if it is not an integer
        if b % 1 == 0.0:
            base = '%d' % b
        else:
            base = '%s' % b

        if abs(fx) < min_exp:
            return r'\$%s%g\$' % (sign_string, x)
        elif not is_x_decade:
            return self._non_decade_format(sign_string, base, fx)
        else:
            return r'\$%s%s^{%d}\$' % (sign_string, base, fx)


@dman.storable(name='__plt_svg')
class SVGFigure:
    __ext__ = '.svg'

    def __init__(self, fig: plt.Figure, generate_tex: bool = False, **kwargs):
        self.fig = fig
        self.kwargs = kwargs
        self.generate_tex = generate_tex
    
    def _generate_tex(self, base: str):
        try:
            subprocess.run([
                'inkscape', 
                '-D', 
                '-z', 
                f'--file={base}.svg', 
                f'--export-pdf={base}.pdf', 
                '--export-latex'
            ])
        except FileNotFoundError:
            dman.log.warning('Inkscape is not installed on this platform.')
        
        if not os.path.exists(f'{base}.pdf_tex'):
            raise RuntimeError('Inkscape failed to generate tex at "{base}.pdf_tex".')
        
        with open(f'{base}.pdf_tex', 'r') as f:
            content = f.readlines()

        idx1 = next(i for i, e in enumerate(content) if 'makeatletter' in e)
        idx2 = next(i for i, e in enumerate(content) if 'makeatother' in e)
        pre, post = content[:idx1+1], content[idx2:]
        commands = ''.join(content[idx1+1:idx2])

        # add svgfont command
        commands = textwrap.dedent(commands)
        commands = commands.split('\n')[:-1]
        msg = textwrap.dedent('''\
        \\ifx\\svgfont\\undefined%
        \\global\\let\\svgfont\\footnotesize
        \\fi%    
        ''')
        commands.extend(msg.split('\n'))
        commands = textwrap.indent('\n'.join(commands), prefix=' '*2)
        commands = [commands]

        pattern = re.compile(r'(\\begin\{tabular\}\[t]\{.\})(.*?)(\\end\{tabular\})')
        replace = r'\g<1>\\svgfont{}\g<2>\g<3>'
        for i, line in enumerate(post):
            post[i] = pattern.sub(replace, line)

        with open(f'{base}.tex', 'w') as f:
            f.writelines(pre+commands+post)
        os.remove(f'{base}.pdf_tex')

    def __write__(self, path: str):
        if self.generate_tex:
            keys = ['svg.fonttype', 'mathtext.default', 'text.usetex']
            old = {key: plt.rcParams[key] for key in keys}
            plt.rcParams['svg.fonttype'] = 'none'
            plt.rcParams['mathtext.default'] = 'regular'
            plt.rcParams['text.usetex'] = False        

            changes = []
            for ax in self.fig.get_axes():
                for cax in [ax.xaxis, ax.yaxis]:
                    fmt = cax.get_major_formatter()
                    if isinstance(fmt, matplotlib.ticker.LogFormatter):
                        cax.set_major_formatter(SVGLogFormatter())
                        changes.append((cax, fmt))
                
        self.fig.savefig(path, **self.kwargs)

        if self.generate_tex:
            # TODO has no effect
            plt.rcParams.update(old)
            for cax, fmt in changes:
                cax.set_major_formatter(fmt)
            
            self._generate_tex(os.path.splitext(path)[0])

    @classmethod
    def __read__(self):
        raise RuntimeError('Cannot read PrintFigure objects.')


@dman.storable(name='__plt_pdf')
class PdfFigure:
    __ext__ = '.pdf'

    def __init__(self, fig: Union[List[plt.Figure], plt.Figure] = None, path: str = None, **kwargs):
        if not isinstance(fig, (type(None), list)):
            fig = [fig]
        self.fig = fig
        self.path = path
        self.kwargs = kwargs

    def __write__(self, path: str):
        if self.fig is not None:
            pdf = PdfPages(path)
            for fig in self.fig:
                pdf.savefig(fig, **self.kwargs)
            pdf.close()
        elif os.path.exists(self.path) and self.path != path:
            shutil.copyfile(self.path, path)

        self.path = path
            
    @classmethod
    def __read__(cls, path: str):
        return cls(path=path)


@dman.storable(name='__plt_pkl')
class PklFigure:
    __ext__ = '.fig'

    def __init__(self, fig: Union[List[plt.Figure], plt.Figure]):
        if not isinstance(fig, list):
            fig = [fig]
        self.fig = fig
    
    def __write__(self, path: str):
        with open(path, 'wb') as f:
            pickle.dump(self.fig, f)
    
    @classmethod
    def __read__(cls, path: str):
        with open(path, 'rb') as f:
            fig = pickle.load(f)
        return cls(fig=fig)



try:
    import tikzplotlib
    @dman.storable(name='__plt_tikz')
    class TexFigure:
        __ext__ = '.tex'

        def __init__(self, fig: plt.Figure = None, path: str = None):
            self.path = path
            self.fig = fig
        
        def __write__(self, path: str):
            if self.fig is not None:
                # tikzplotlib.clean_figure()
                tikzplotlib.save(
                    figure=self.fig,
                    filepath=path,
                    axis_width='\linewidth',
                    externalize_tables=False,
                    strict=False,
                    include_disclaimer=False,
                    standalone=False
                )
            elif self.path != path:
                shutil.copyfile()
            self.path = path
            
        @classmethod
        def __read__(cls, path: str):
            return cls(path=path)
except ImportError as e: ...


@dman.modelclass(name='__plt_t_figure')
class t_Figure:
    pdf: PdfFigure = dman.recordfield(stem='preview')
    fig: PklFigure = dman.recordfield(stem='content')

    def __post_init__(self):
        self.pdf.fig = self.fig.fig

    @classmethod
    def __convert__(cls, other: 'Figure'):
        return cls(pdf=PdfFigure(other.fig), fig=PklFigure(other.fig))


@dman.storable(name='__plt_figure')
@dman.serializable(name='__plt_figure', template=t_Figure)
class Figure:
    def __init__(self, fig: Union[List[plt.Figure], plt.Figure]):
        if not isinstance(fig, list):
            fig = [fig]
        self.fig = fig

    @classmethod
    def __convert__(cls, other: t_Figure):
        return cls(fig=other.fig)
        

def main():
    # create the figures and save to file ......................................
    f1 = plt.figure()
    plt.plot(np.linspace(0, 1, 100), np.linspace(0, 1, 100)**2)

    f2 = plt.figure()
    plt.plot(np.linspace(0, 1, 100), (np.linspace(0, 1, 100)-0.5)**2)

    dman.save('fig', Figure([f1, f2]))
    plt.close('all')

    # load figures from file (can be executed from different script) ...........
    fig = dman.load('fig')
    print(fig)
    plt.show()


if __name__ == '__main__':
    main()


from dman.core.serializables import serializable, register_serializable
from dman.core.storables import storable
from dman.model.modelclasses import register_preset
from dman.utils import sjson
import dman.model.modelclasses

from typing import Type, Union
import numpy as np


class _typed_array(type):
    __cached__ = dict()

    def __getitem__(cls, tp: Type):
        res = _typed_array.__cached__.get((cls, tp), None)
        if res is not None:
            return res

        name = f"{cls.__name__}[{tp.__name__}]"
        res = _typed_array.__new__(
            _typed_array, name, (cls,), {"__qualname__": name, "__module__": __name__}
        )
        register_preset(
            res,
            lambda arg: arg.view(res).astype(tp)
            if isinstance(arg, np.ndarray)
            else arg,
        )
        _typed_array.__cached__[(cls, tp)] = res
        return res


@storable(name="_num__barray")
class barray(np.ndarray, metaclass=_typed_array):
    __ext__ = ".npy"

    def __write__(self, path):
        with open(path, "wb") as f:
            np.save(f, self)

    @classmethod
    def __read__(cls, path):
        with open(path, "rb") as f:
            res: np.ndarray = np.load(f)
            return res.view(cls)


@serializable(name="_num__sarray")
class sarray(np.ndarray, metaclass=_typed_array):
    def __serialize__(self):
        if self.ndim == 0:
            return sjson.dumps(float(self))
        if self.size == 0:
            return {"shape": sjson.dumps(self.shape)}
        elif self.ndim == 1:
            return sjson.dumps(self.tolist())
        return [sarray.__serialize__(a) for a in self]

    @classmethod
    def __deserialize__(cls, obj: Union[list, str]):
        if isinstance(obj, str):
            return np.asarray(sjson.loads(obj)).view(cls)
        if isinstance(obj, dict):
            return np.zeros(sjson.loads(obj.get("shape")))
        if isinstance(obj, (float, int)):
            return np.asarray(obj).view(cls)
        return np.asarray([sarray.__deserialize__(a) for a in obj]).view(cls)


@serializable(name="_num__carray")
class carray(np.ndarray, metaclass=_typed_array):
    def __eq__(self, other):
        return np.array_equal(self, other)

    def __serialize__(self):
        return self.view(sarray).__serialize__()

    @classmethod
    def __deserialize__(cls, obj: Union[list, str]):
        return sarray.__deserialize__(obj).view(cls)


register_preset(
    barray, lambda arg: arg.view(barray) if isinstance(arg, np.ndarray) else arg
)
register_preset(
    carray, lambda arg: arg.view(carray) if isinstance(arg, np.ndarray) else arg
)
register_preset(
    sarray, lambda arg: arg.view(sarray) if isinstance(arg, np.ndarray) else arg
)


from dman.utils.sjson import register_atomic_alias

register_atomic_alias(np.int32, int)
register_atomic_alias(np.int64, int)
register_atomic_alias(np.float32, float)
register_atomic_alias(np.float64, float)


register_serializable(
    "_num__ndarray",
    np.ndarray,
    serialize=lambda ser: ser.view(sarray).__serialize__(),
    deserialize=lambda ser: sarray.__deserialize__(ser).view(np.ndarray),
)


if __name__ == "__main__":
    import dman

    ser = dman.save("array", np.eye(10))
    print(dman.sjson.dumps(ser, indent=4))
    print(dman.load("array"))



from dataclasses import dataclass
from dman.core import serializables
from dman.core import path
from dman.model import modelclasses
from dman.utils.smartdataclasses import configclass

@configclass
class Config:
    serialize: serializables.Config = None
    store: path.Config = None
    model: modelclasses.Config = None


params = Config(
    serializables.config, 
    path.config, 
    modelclasses.config
)


"""
Contains utilities for creating serializable types and to serialize and 
deserialize them into objects that can be encoded through json.
"""


from dataclasses import MISSING, dataclass, fields, is_dataclass, asdict
import inspect
import sys
from typing import Any, Callable, Optional, Sequence, Type


from dman.utils import sjson
from dman.core import log
from dman.core.errors import Trace, Stack, Frame, BaseInvalid, ExcInvalid
from dman.utils.smartdataclasses import configclass
import textwrap
from contextlib import suppress

from enum import Enum


SER_TYPE = "_ser__type"
SER_CONTENT = "_ser__content"
SERIALIZE = "__serialize__"
DESERIALIZE = "__deserialize__"
NO_SERIALIZE = "__no_serialize__"
CONVERT = "__convert__"
DECONVERT = "__de_convert__"


__serializable_types = dict()
__custom_serializable = dict()


@configclass
class Config:
    """Configuration for serializables.

        This class has a global instance that can be accessed as follows:

        >>> dman.core.serializables.config.validate = True
        >>> dman.params.serialize.validate = True  # equivalent

    Args:
        validate (bool, optional): Cancel serialization when an invalid object is encountered. Defaults to False.
    """

    validate: bool = False


config = Config()


def compare_type(base: Type, check: Sequence[Type]):
    """Compare two types, supporting generic types.

    The signature operates similarly to ``isinstance``.
    """
    if not isinstance(check, Sequence):
        check = (check,)
    if base in check:
        return True
    base = getattr(base, "__origin__", MISSING)
    return base is not MISSING and base in check
    

def get_type_name(ser_type):
    """Get the class name of an instance or type."""
    name = getattr(ser_type, "__name__", None)
    if name is None:
        name = getattr(ser_type, "name", str(ser_type))
    return name


def ser_type2str(obj):
    """
    Returns the serialize type string of an object or class."""
    return getattr(obj, SER_TYPE, None)


def ser_str2type(ser):
    """
    Returns the class represented by a serialize type string."""
    return __serializable_types.get(ser, None)


def is_serializable(ser):
    """Check if an object is a serializable type."""
    if _Instance.__convert__(ser) is not None:
        return True
    if is_serializable_type_str(getattr(ser, SER_TYPE, None)):
        return True
    if isinstance(ser, (list, dict, tuple)):
        return True
    if sjson.atomic_type(ser):
        return True
    return type(ser) in __custom_serializable
    

def is_serializable_type_str(ser: str):
    """Check if a string is the name of a registered serializable type."""
    return ser is not None and ser in __serializable_types


def is_deserializable(serialized: Any):
    """Check if an object has been produced through serialization.

    If the serialized object is a dictionary, then it should include
    the serializable type string.
    """
    if sjson.atomic_type(serialized) or isinstance(serialized, (list, tuple)):
        return True
    if not isinstance(serialized, dict):
        return False
    ser_type = serialized.get(SER_TYPE, None)
    if ser_type:
        return is_serializable_type_str(ser_type)
    return False


def register_serializable(
    name: str,
    cls: Type[Any],
    serialize: Callable[[Any, Optional["BaseContext"]], Any] = None,
    deserialize: Callable[[Any, Optional["BaseContext"]], Any] = None,
):
    """Register a class as a serializable type with a given name.

        If the serialize and deserialize methods are not provided then the class
        should have a `__serialize__` and `__deserialize__` method defined.
        In this case however you should prefer using the :func:``serializable``
        decorator instead.

    Args:
        name (str): Name of the serializable
        cls (Type[Any]): Class to register
        serialize (Callable[[Any, Optional[BaseContext]], Any], optional): Serialize method. Defaults to None.
        deserialize (Callable[[Any, Optional[BaseContext]], Any], optional): Deserialize method. Defaults to None.

    Example:
        >>> class Base: ...
        >>> register_serializable('base', Base, lambda obj: '<base>', lambda ser: Base())
    """
    __serializable_types[name] = cls
    if serialize is not None and deserialize is not None:
        __custom_serializable[cls] = (name, serialize, deserialize)


def serializable_types():
    """Return the serializable types dictionary."""
    return __serializable_types


def get_custom_serializable(cls: Type[Any], default=None):
    """Get the signature of a custom serializable object."""
    return __custom_serializable.get(cls, default)

def serialize(obj, context: "BaseContext" = None, content_only: bool = False):
    """Serialize a serializable object.

        The context can be used to store the current directory, logging
        and custom serialization behavior. Serializable objects are created
        using :func:`serializable` or alternatively using :func:`register_serializable`.

        By default, any object that can be handled by json is serializable.
        That is ``str``, ``int``, ``float``, ``None``, ``list``, ``tuple`` and ``dict``.

    Args:
        obj: The object to serialize
        context (BaseContext, optional): The serialization context. Defaults to None.
        content_only (bool, optional): Do not include type information when true.
            Results in more compact serialization. Defaults to False.

    Example:
        >>> @serializable
        >>> @dataclass
        >>> class Base:
        >>>     value: str
        >>> ser = serialize([Base('content'), 'message'])
        >>> print(sjson.dumps(ser, indent=4))
        [
            {
                "_ser__type": "Base",
                "_ser__content": {
                    "value": "content"
                }
            },
            "message"
        ]
        >>> ser = serialize(Base('content'), content_only=True)
        >>> print(sjson.dumps(ser, indent=4))
        {
            "value": "content"
        }
    """

    if context is None:
        context = BaseContext()
    return context.serialize(obj, content_only=content_only)


def deserialize(serialized, context: "BaseContext" = None, expected=None):
    """Deserialize an object produced through serialization.

        The context can be used to store the current directory, logging
        and custom serialization behavior. Deserializable objects are created
        using :func:`serializable` or alternatively using :func:`register_serializable`.

        By default, any object that can be handled by json is deserializable.
        That is ``str``, ``int``, ``float``, ``None``, ``list``, ``tuple`` and ``dict``.

    Args:
        serialized: The object to deserialize.
        context (BaseContext, optional): _description_. Defaults to None.
        expected (_type_, optional): Class or string representing the expected type. If set to
            None the type is received from the dictionary. Defaults to None.

    Example:
        >>> @serializable(name='_base')
        >>> @dataclass
        >>> class Base:
        >>>     value: str
        >>> ser = serialize(Base('content'))
        >>> print(deserialize(ser).value)
        content
        >>> ser = serialize(Base('content'), content_only=True)
        >>> print(deserialize(ser, expected=Base).value)
        content
        >>> print(deserialize(ser, expected='_base').value)
        content
    """

    if context is None:
        context = BaseContext()
    return context.deserialize(serialized, expected)


def _serialize__dataclass(self, context: "BaseContext" = None):
    serialized = dict()
    for f in fields(self):
        if f.name not in getattr(self, NO_SERIALIZE, []):
            value = getattr(self, f.name)
            serialized[f.name] = serialize(value, context)

    return serialized


@classmethod
def _deserialize__dataclass(cls, serialized: dict, context: "BaseContext"):
    res = dict()
    for k, v in serialized.items():
        res[k] = deserialize(v, context)

    return cls(**res)


def _serialize__enum(self):
    return str(self)


@classmethod
def _deserialize__enum(cls, serialized: str):
    *_, name = serialized.split(".")
    return cls[name]


@classmethod
def _default__convert(cls, base):
    return cls(**asdict(base))


def _serialize__template(template: Any, cls: any):
    convert = getattr(template, CONVERT, None)
    if convert is None:
        if not is_dataclass(cls):
            raise ValueError(
                f'Serializable should either be a dataclass or should have a "{CONVERT}" method specified.'
            )
        setattr(template, CONVERT, _default__convert)
        convert = getattr(template, CONVERT, None)

    def __serialize__(self, context: BaseContext = None):
        return serialize(convert(self), context, content_only=True)

    return __serialize__


def _deserialize__template(template: Any):
    @classmethod
    def __deserialize__(cls, ser, context: BaseContext = None):
        convert = getattr(cls, CONVERT, None)
        if convert is None:
            convert = getattr(template, DECONVERT, None)
        if convert is None:
            convert = lambda x: x
        return convert(deserialize(ser, context, expected=template))

    return __deserialize__


def serializable(cls=None, /, *, name: str = None, template: Any = None):
    """Make a class serializable.

        Returns the same class as was passed in and the class is registered as a
        serializable type. Serialization and Deserialization methods are added
        automatically if cls is a dataclass or an Enum. Otherwise a
        ``__serialize__`` and ``__deserialize__`` method should be provided
        for serialization. If these are not provided conversion will fail.

        See :ref:`sphx_glr_gallery_fundamentals_example0_serializables.py` for
        detailed examples on how to create serializable types.

    Args:
        cls (Any): The class to process.
        name (str, optional): The name of the serializable type. Defaults to None.
        template (Any, optional): Template class to use during serialization. Defaults to None.
    """

    def wrap(cls):
        local_name = name
        if local_name is None:
            local_name = getattr(cls, "__name__")
        setattr(cls, SER_TYPE, local_name)
        register_serializable(local_name, cls)

        if template is not None:
            if not hasattr(cls, CONVERT):
                if not is_dataclass(template):
                    raise ValueError(
                        f'Template should be either a dataclass or should have a "{CONVERT}" method defined.'
                    )
                if not hasattr(template, DECONVERT):
                    setattr(cls, CONVERT, _default__convert)
            if getattr(cls, SERIALIZE, None) is None:
                setattr(cls, SERIALIZE, _serialize__template(template, cls))
            if getattr(cls, DESERIALIZE, None) is None:
                setattr(cls, DESERIALIZE, _deserialize__template(template))

        elif is_dataclass(cls):
            if getattr(cls, SERIALIZE, None) is None:
                setattr(cls, SERIALIZE, _serialize__dataclass)
            if getattr(cls, DESERIALIZE, None) is None:
                setattr(cls, DESERIALIZE, _deserialize__dataclass)

        elif issubclass(cls, Enum):
            if getattr(cls, SERIALIZE, None) is None:
                setattr(cls, SERIALIZE, _serialize__enum)
            if getattr(cls, DESERIALIZE, None) is None:
                setattr(cls, DESERIALIZE, _deserialize__enum)

        if not hasattr(cls, SERIALIZE) or not hasattr(cls, DESERIALIZE):
            raise ValueError(
                f"Class {cls} could not be made serializable. Provide a manual definition of a `__serialize__` and `__deserialize__` method."
            )

        return cls

    # See if we're being called as @serializable or @serializable().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @serializable without parens.
    return wrap(cls)


serializable(Frame, name="__frame")
serializable(Stack, name="__stack")
serializable(Trace, name="__trace")


@serializable(name='__instance')
class _Instance:
    registered = {}
    def __init__(self, name: str):
        if name not in self.__class__.registered:
            raise ValueError('Created an instance tracker for an unregistered name')
        self.name = name
    
    @classmethod
    def __convert__(cls, inst: Any):
        for k, v in cls.registered.items():
            if v is inst:
                return cls(k)
        return None

    def __serialize__(self):
        return self.name
    
    @classmethod
    def __deserialize__(cls, ser):
        return cls.registered[ser]


def register_instance(inst: Any = None, /, *, name: str = MISSING):
    """Register instance for serialization

    Args:
        inst (Any): Instance to serialize. Defaults to None.
        name (str): Name of instances.
    
    Example:
        >>> @dman.register_instance(name='ell1')
        >>> def ell1(x, y):
        >>>     return abs(x) + abs(y)

        >>> dman.register_instance(ell1, name='ell1')
    """
    if name is MISSING:
        raise TypeError("registered_instance() missing 1 required keyword-only argument: 'name'")
    
    def wrap(inst):
        old = _Instance.registered.get(name, None)
        if old is not None:
            log.warning(f'Overwrote registered instance with name {name} from {old.__repr__()} to {inst.__repr__()}.')
        _Instance.registered[name] = inst
        return inst
    if inst is None:
        return wrap
    return wrap(inst)
        


@serializable(name="__unserializable")
class Unserializable(BaseInvalid):
    """Represents an object that could not be serialized."""

    ...


@serializable(name="__exc_unserializable")
class ExcUnserializable(ExcInvalid):
    """Represents an object that could not be serialized due to an exception."""

    ...


@serializable(name="__undeserializable")
@dataclass(repr=False)
class Undeserializable(BaseInvalid):
    """Represents an object that could not be deserialized."""

    serialized: dict = None  #: The serialized object.
    expected: str = None  #: The expected serializable type of the object.

    def format(self):
        yield from super().format()
        if self.serialized:
            yield "\n\nSerialized\n"
            yield textwrap.indent(sjson.dumps(self.serialized, indent=4), " " * 0)

    def __serialize__(self):
        res = super().__serialize__()
        if self.serialized is not None:
            res["serialized"] = self.serialized
        if self.expected is not None:
            res["ser_type"] = self.expected
        return res

    @classmethod
    def __deserialize__(cls, serialized: dict):
        res = cls(**deserialize(serialized))
        if res.serialized is not None:
            return deserialize(res.serialized, res.expected)
        return res


@serializable(name="__exc_undeserializable")
@dataclass(repr=False)
class ExcUndeserializable(ExcInvalid):
    """Represents an object that could not be deserialized due to an exception."""

    serialized: dict
    expected: str

    def format(self):
        yield from super().format()
        if self.serialized:
            yield "\n\nSerialized\n"
            yield textwrap.indent(sjson.dumps(self.serialized, indent=4), " " * 0)

    def __serialize__(self):
        res = super().__serialize__()
        if self.serialized is not None:
            res["serialized"] = self.serialized
        if self.expected is not None:
            res["expected"] = self.expected
        return res

    @classmethod
    def __deserialize__(cls, serialized: dict, context):
        res = cls(**serialized)
        if res.serialized is not None:
            return deserialize(res.serialized, context, res.expected)
        return res


def isvalid(obj):
    """
    Check if an object is valid
    """
    if isinstance(obj, dict):
        return not issubclass(ser_str2type(obj.get(SER_TYPE, None)), BaseInvalid)
    return not isinstance(obj, BaseInvalid)


class SerializationError(Exception):
    """
    Serialization Error raised when an object could not be serialized.
    """

    pass


class ValidationError(SerializationError):
    """
    Validation Error raised when an object could not be validated.
    """

    pass


def validate(obj, msg: str = "Could not validate object"):
    """Check if the object is valid, raise a ValidationError otherwise.

    Args:
        obj: The object to validate
        msg (str, optional): An optional message set in the Validation Error.
            Defaults to 'Could not validate object'.

    Raises:
        ValidationError: The object is not valid.
    """

    if isvalid(obj):
        return
    raise ValidationError(msg, str(obj))


def _call_optional_context(
    method, *args, context=None, exc_type: Type[ExcInvalid] = None, **kwargs
):
    try:
        sig = inspect.signature(method)
        if len(sig.parameters) == len(args):
            return method(*args)
        if len(sig.parameters) == len(args) + 1:
            return method(*args, BaseContext() if context is None else context)
        raise TypeError(
            f"Expected method that takes {len(args)} or {len(args)+1} positional arguments but got {len(sig.parameters)}."
        )
    except SerializationError:
        raise
    except Exception as e:
        if exc_type is None:
            raise e
        return exc_type.from_exception(*sys.exc_info(), **kwargs, ignore=1)


class BaseContext:
    """
    The basic interface for serialization contexts.

    It is passed hierarchically through the hierarchy of objects that
    are being serialized and describes how the serialization should be completed.
    For :class:`dman.model.record.Context` for more information.
    """

    def process_invalid(self, msg: str, obj: BaseInvalid):
        if isinstance(obj, ExcInvalid):
            log.logger.warning(
                msg + "\n" + "".join(BaseInvalid.format(obj)),
                "context",
                obj.trace,
                stacklevel=2,
            )
        else:
            log.logger.warning(msg + "\n" + str(obj), "context", stacklevel=2)

        if config.validate:
            raise ValidationError(
                msg
                + "\n\nDescription:\n"
                + f"[{log.logger.format_stack()}] "
                + str(obj)
            )

    def serialize(self, obj, content_only: bool = False):
        """Serialize a serializable object.

            The context can be used to store the current directory, logging
            and custom serialization behavior. Serializable objects are created
            using :func:`serializable` or alternatively using :func:`register_serializable`.

            By default, any object that can be handled by json is serializable.
            That is ``str``, ``int``, ``float``, ``None``, ``list``, ``tuple`` and ``dict``.

        Args:
            obj: The object to serialize
            context (BaseContext, optional): The serialization context. Defaults to None.
            content_only (bool, optional): Do not include type information when true.
                Results in more compact serialization. Defaults to False.

        See also: :func:`serialize`.
        """
        # check whether the object a registered instance. If so, convert it.
        inst = _Instance.__convert__(obj)
        if inst is not None:
            obj = inst

        if sjson.atomic_type(obj):
            return obj

        if isinstance(obj, (list, tuple)):
            with log.layer(f"list({len(obj)})", "serializing", owner=list):
                return self.serialize_list(obj)

        if isinstance(obj, dict):
            with log.layer(f"dict({len(obj)})", "serializing", owner=dict):
                return self.serialize_dict(obj)

        # if isinstance(ser, BaseInvalid):
        #     self._process_invalid('Invalid object encountered during serialization.', ser)

        with log.layer(f"{type(obj).__name__}", "serializing", owner=obj):
            ser_type, content = self.serialize_object(obj)
            if isinstance(content, BaseInvalid):
                self.process_invalid(
                    "Serialization resulted in an invalid object.", content
                )
                ser_type, content = self.serialize_object(content)

        if content_only:
            return content
        return {SER_TYPE: ser_type, SER_CONTENT: content}

    def serialize_object(self, obj):
        """Serialize a generic serializable object."""
        ser_type, ser_method, _ = get_custom_serializable(type(obj), (None, None, None))
        if ser_type is None:
            ser_type = getattr(obj, SER_TYPE, None)
            if not is_serializable_type_str(ser_type):
                return None, Unserializable(
                    type=type(obj).__name__,
                    info=f"Unserializable type: {type(obj).__name__}.",
                )
            ser_method = getattr(obj, SERIALIZE, None)
            return ser_type, _call_optional_context(
                ser_method,
                context=self,
                exc_type=ExcUnserializable,
                type=type(obj).__name__,
                info="Error during serialization:",
            )
        else:
            return ser_type, _call_optional_context(
                ser_method,
                obj,
                context=self,
                exc_type=ExcUnserializable,
                type=type(obj).__name__,
                info="Error during serialization:",
            )

    def serialize_list(self, obj: list):
        """Serialize a list."""
        return [itm if sjson.atomic_type(itm) else self.serialize(itm) for itm in obj]

    def serialize_dict(self, obj: dict):
        """Serialize a dictionary."""
        return {self.serialize(k): self.serialize(v) for k, v in obj.items()}

    def deserialize(self, serialized, ser_type=None):
        """Deserialize an object produced through serialization.

            The context can be used to store the current directory, logging
            and custom serialization behavior. Deserializable objects are created
            using :func:`serializable` or alternatively using :func:`register_serializable`.

            By default, any object that can be handled by json is deserializable.
            That is ``str``, ``int``, ``float``, ``None``, ``list``, ``tuple`` and ``dict``.

        Args:
            serialized: The object to deserialize.
            context (BaseContext, optional): _description_. Defaults to None.
            expected (_type_, optional): Class or string representing the expected type. If set to
                None the type is received from the dictionary. Defaults to None.

            See also: :func:`deserialize`.
        """
        res = self._deserialize_inner(serialized, ser_type)
        if isinstance(res, BaseInvalid):
            self.process_invalid("An error occurred during deserialization:", res)
        return res

    def _deserialize_inner(self, serialized, expected=None):
        if serialized is None:
            return None

        if expected is None:
            if isinstance(serialized, (list, tuple)):
                with log.layer(f"list({len(serialized)})", "deserializing", owner=list):
                    return self.deserialize_list(serialized, list)

            if sjson.atomic_type(serialized):
                return self.deserialize_atomic(serialized, type(serialized))

            if not isinstance(serialized, dict):
                return Undeserializable(
                    type=expected,
                    info=f"Unexpected type for serialized: {type(serialized)}. Expected either list, tuple, atomic type or dict.",
                    serialized=serialized,
                )

            expected = serialized.get(SER_TYPE, MISSING)
            if expected is MISSING:
                with log.layer(f"dict({len(serialized)})", "deserializing", owner=dict):
                    return self.deserialize_dict(serialized, dict)
            else:
                serialized = serialized.get(SER_CONTENT, {})
                ser_type_get = ser_str2type(expected)
                if ser_type_get is None:
                    return Undeserializable(
                        type=ser_type_get,
                        info=f"Unregistered type stored in serialized.",
                        serialized=serialized,
                        expected=expected,
                    )

                _ser_name = get_type_name(ser_type_get)
                with log.layer(f"{_ser_name}", "deserializing", owner=ser_type_get):
                    return self.deserialize_object(serialized, ser_type_get)

        if compare_type(expected, dict):
            with log.layer(f"dict({len(serialized)})", "deserializing", owner=dict):
                return self.deserialize_dict(serialized, dict)

        if compare_type(expected, (list, tuple)):
            with log.layer(f"list({len(serialized)})", "deserializing", owner=list):
                return self.deserialize_list(serialized, list)

        if isinstance(expected, str):
            ser_type_get = ser_str2type(expected)
            if ser_type_get is None:
                return Undeserializable(
                    type=ser_type_get,
                    info=f"Unregistered type provided as argument.",
                    serialized=serialized,
                    expected=expected,
                )
            expected = ser_type_get

        if expected in sjson.atomic_types:
            return self.deserialize_atomic(serialized, expected)

        _ser_name = get_type_name(expected)
        with log.layer(f"{_ser_name}", "deserializing", owner=expected):
            return self.deserialize_object(serialized, expected)

    def deserialize_object(self, serialized, expected):
        """Deserialize an object of an expected type."""
        _, _, ser_method = get_custom_serializable(expected, (None, None, None))
        if ser_method is None:
            ser_method = getattr(expected, DESERIALIZE, None)
        if ser_method is None:
            return Undeserializable(
                get_type_name(expected), 
                f"Could not recover __deserialize__ method for type '{get_type_name(expected)}'.", 
                serialized
            )
        return _call_optional_context(
            ser_method,
            serialized,
            context=self,
            exc_type=ExcUndeserializable,
            type=ser_type2str(expected),
            info=f"Exception encountered while deserializing {expected}:",
            serialized=serialized,
            expected=ser_type2str(expected),
        )

    def deserialize_atomic(self, serialized, expected):
        """Deserialize an atomic object of an expected type"""
        with suppress(ValueError):
            return expected(serialized)
        if expected is not type(serialized):
            return Undeserializable(
                type=expected,
                info=f"Specified type {expected}, but got {type(serialized)}.",
                serialized=serialized,
            )
        return serialized

    def deserialize_list(self, serialized, expected):
        res = [
            itm if sjson.atomic_type(itm) else self.deserialize(itm)
            for itm in serialized
        ]
        if expected is list:
            return res
        return expected(res)

    def deserialize_dict(self, serialized: dict, expected):
        res = {self.deserialize(k): self.deserialize(v) for k, v in serialized.items()}
        if expected is dict:
            return res
        return expected(res)



import argparse
from dman.core.path import get_root_path



def init_dman():
    get_root_path(create=True)

def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        title='commands', 
        description='dman supports the actions listed below', 
        help='options'
    )

    # == init command ==============================================
    init_parser = subparsers.add_parser(
        'init', 
        help='initialize dman in this folder'
    )
    init_parser.set_defaults(
        execute=lambda _: init_dman()
    )

    # == parse arguments =======================================================
    args = parser.parse_args()
    execute = getattr(args, 'execute', None)
    if execute:
        execute(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()